import os
import re
import sys
import logging
import warnings
import idlib
from idlib.utils import log as _ilog
from augpathlib.utils import log as _alog
from pyontutils.utils import (makeSimpleLogger,
                              python_identifier,  # FIXME update imports
                              TZLOCAL,
                              utcnowtz,
                              isoformat,
                              isoformat_safe,
                              timeformat_friendly)
from . import exceptions as exc
from .config import auth

_find_command = 'gfind' if sys.platform == 'darwin' else 'find'

log = makeSimpleLogger('sparcur')
logd = log.getChild('data')
loge = log.getChild('export')

# set augpathlib log format to pyontutils (also sets all child logs)
_alog.removeHandler(_alog.handlers[0])
_alog.addHandler(log.handlers[0])
# idlib logs TODO move to pyontutils probably?
_ilog.removeHandler(_alog.handlers[0])
_ilog.addHandler(log.handlers[0])

# silence known warnings on pypy3
if hasattr(sys, 'pypy_version_info'):
    warnings.filterwarnings('ignore', module='.+protobuf.+')

__type_registry = {None: None}
def register_type(cls, type_name):
    if type_name in __type_registry:
        if __type_registry[type_name] is cls:
            # better to do this check here than to force
            # all callers to check for themselves which
            # can fail if two separate systems try to
            # register the same type
            return

        raise ValueError(f'Cannot map {cls} to {type_name}. '
                         'Type already present! '
                         f'{type_name} -> {__type_registry[type_name]}')

    __type_registry[type_name] = cls


def register_all_types():
    # as a side effect this registers idlib streams and OntTerm
    # sigh doing anything in the top level of python :/
    import sparcur.core
    import sparcur.paths  # also a top level registration

    # this is not done at top level because it is quite slow
    from pysercomb.pyr import units as pyru
    [register_type(c, c.tag)
     for c in (pyru._Quant, pyru.Range, pyru.Approximately)]


class IdentityJsonType:
    """ use to register types that should not be recursed upon e.g.
        because they contain external use of type that does not align
        with our usage """

    @classmethod
    def fromJson(cls, blob):
        return blob


def fromJson(blob, *,
             # nct is a hack around the fact that external
             # data can provide a column called type and we
             # do not currently have a way to indicate that
             # fromJson should treat those subtrees specially
             _no_convert_type=False,
             _no_convert_type_keys=('subjects', 'samples')):
    def nitr(value):
        try:
            return value not in __type_registry
        except TypeError as e:
            log.critical(e)
            return False

    if isinstance(blob, dict):
        if 'type' in blob:
            t = blob['type']

            if t == 'identifier':
                type_name = blob['system']
            elif t in ('quantity', 'range'):
                type_name = t
            elif _no_convert_type:
                type_name = None
            elif nitr(t):
                breakpoint()
                # XXX FIXME should this be fatal ??? we have cases where
                # it is impossible to avoid because a user will use type
                # as a column header and then we are toast if we also want
                # to set the type of that object to convert back ... what
                # happens if we want to have subject metadata as a class?
                # do we just ... hope for the best that there are no collisions
                # and make collisions non-fatal ???
                msg = (f'TODO fromJson for type {t} '
                       f'currently not implemented\n{blob}')
                log.critical(msg)
                if False:  # set true if debugging when implementing new types
                    raise NotImplementedError(msg)
                type_name = None
            else:
                type_name = t

            cls = __type_registry[type_name]
            if cls is not None:
                return cls.fromJson(blob)

        return {k: v
                if k == 'errors' or k.endswith('_errors') else
                fromJson(v, _no_convert_type=(_no_convert_type or
                                              type(v) == list and
                                              k in _no_convert_type_keys))
                for k, v in blob.items()}

    elif isinstance(blob, list):
        return [fromJson(_, _no_convert_type=_no_convert_type) for _ in blob]
    else:
        return blob


def path_irs(*paths_or_strings):
    """Given one or more paths pointing to sparcur export
    json yield the python internal representation."""
    # TODO support for urls
    import json
    register_all_types()

    for path_or_string in paths_or_strings:
        with open(path_or_string) as f:
            blob = json.load(f)

        yield fromJson(blob)


def path_ir(path_or_string):
    """Given a path or string return the sparcur python ir."""
    return next(path_irs(path_or_string))


def render_manifest(rows):
    # FIXME checksum cypher
    return [[filename.as_posix(),
             isoformat(timestamp),
             desc,
             ft,
             at,
             checksum.hex() if checksum else checksum,]
            for filename, timestamp, desc, ft, at, checksum in rows]


def write_manifests(*args, parents=None, parents_rows=None, suffix='.csv',
                    include_directories=False,):
    header = ('filename', 'timestamp', 'description',  # FIXME no hardcode
              'file type', 'additional types', 'checksum')
    if parents is None and parents_rows is None:
        raise TypeError('one of parents or parents_rows is required')
    elif parents and parents_rows:
        raise TypeError('at most one of parents or parents_rows is allowed')

    if parents_rows:
        parents = [p for p, r in parents_rows]

    existing = []
    for parent in parents:
        manifest = parent / f'manifest{suffix}'
        if manifest.exists():
            existing.append(manifest)

    if existing:  # TODO overwrite etc.
        # FIXME TODO relative to some reference point
        msg = f'Existing manifest files detected not writing!\n{existing}'
        raise ValueError(msg)

    if parents_rows is None:
        parents_rows = [(path, path.generate_manifest())
                        for path in parents]

    manifests_rendered = []
    if suffix == '.csv':  # FIXME deal with different suffixes
        import csv
        paths_rendered = [(path, render_manifest(manifest))
                          for path, manifest in parents_rows]
        for path, rendered in paths_rendered:
            manifest = parent / f'manifest{suffix}'
            manifests_rendered.append((manifest, rendered))
            with open(manifest, 'wt') as f:
                csv.writer(f).writerows([header] + rendered)
    else:
        raise NotImplementedError(f"Don't know how to export {suffix}")

    return manifests_rendered


def expand_label_curie(rows_of_terms):
    return [[value for term in rot for value in
             (term.label if term is not None else '',
              term.curie if term is not None else '')]
            for rot in rows_of_terms]


def levenshteinDistance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = range(len(s1) + 1)
    for i2, c2 in enumerate(s2):
        distances_ = [i2+1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1],
                                           distances[i1 + 1],
                                           distances_[-1])))
        distances = distances_
    return distances[-1]


class GetTimeNow:
    def __init__(self):
        self._start_time = utcnowtz()
        self._start_local_tz = TZLOCAL()  # usually PST PDT

    @property
    def _start_time_local(self):
        return self._start_time.astimezone(self._start_local_tz)

    @property
    def START_TIMESTAMP(self):
        return isoformat(self._start_time)

    @property
    def START_TIMESTAMP_SAFE(self):
        return isoformat_safe(self._start_time)

    @property
    def START_TIMESTAMP_FRIENDLY(self):
        return timeformat_friendly(self._start_time)

    @property
    def START_TIMESTAMP_LOCAL(self):
        return isoformat(self._start_time_local)

    @property
    def START_TIMESTAMP_LOCAL_SAFE(self):
        return isoformat_safe(self._start_time_local)

    @property
    def START_TIMESTAMP_LOCAL_FRIENDLY(self):
        return timeformat_friendly(self._start_time_local)


class SimpleFileHandler:

    _FIRST = object()

    def __init__(self, log_file_path, *logs, mimic=_FIRST):
        self.log_file_handler = logging.FileHandler(log_file_path.as_posix())
        if mimic is self._FIRST and logs:
            self.mimic(logs[0])
        elif mimic:
            self.mimic(mimic)

        for log in logs:
            self(log)

    def __call__(self, *logs_to_handle):
        for log in logs_to_handle:
            log.addHandler(self.log_file_handler)

    def mimic(self, log):
        self.log_file_handler.setFormatter(log.handlers[0].formatter)


def silence_loggers(*logs):
    for log in logs:
        parent = log
        while parent:
            [parent.removeHandler(h) for h in parent.handlers]
            parent = parent.parent


def bind_file_handler(log_file):
    # FIXME the this does not work with joblib at the moment
    from idlib.utils import log as idlog
    from protcur.core import log as prlog
    from orthauth.utils import log as oalog
    from ontquery.utils import log as oqlog
    from augpathlib.utils import log as alog
    from pyontutils.utils import log as pylog

    sfh = SimpleFileHandler(log_file, log)
    sfh(alog, idlog, oalog, oqlog, prlog, pylog)


class _log:
    """ logging prevents nice ipython recurions error printing
        so rename this class to log when you need fake logging """
    @staticmethod
    def debug(nothing): pass
    @staticmethod
    def info(nothing): pass
    @staticmethod
    def warning(nothing): print(nothing)
    @staticmethod
    def error(nothing): pass
    @staticmethod
    def critical(nothing): pass


want_prefixes = ('TEMP', 'FMA', 'UBERON', 'PATO', 'NCBITaxon', 'ilxtr',
                 'sparc', 'BIRNLEX', 'tech', 'unit', 'ILX', 'lex',)


def is_list_or_tuple(obj):
    return isinstance(obj, list) or isinstance(obj, tuple)


def symlink_latest(dump_path, path, relative=True):
    """ relative to allow moves of the containing folder
        without breaking links """

    if relative:
        dump_path = dump_path.relative_path_from(path)

    if path.exists():
        if not path.is_symlink():
            raise TypeError(f'Why is {path.name} not a symlink? '
                            f'{path!r}')

        path.unlink()

    path.symlink_to(dump_path)


def _transitive_(path, command):
    safe_command = command + ' -print0'
    with path:
        with os.popen(safe_command) as p:
            string = p.read()

    path_strings = string.split('\x00')
    # XXXXXXXXXXXXXXXXXXX REMINDER THAT THIS IS NOT SORTED
    # https://doi.org/10.1021/acs.orglett.9b03216
    paths = [path / s for s in path_strings if s][1:]  # leave out the parent folder itself
    return paths


def transitive_paths(path, exclude_patterns=tuple()):
    """Fast list of all child directories using unix find."""
    if sys.platform == 'win32':
        # XXX assumes that rchildren already implements exclude patterns
        return list(path.rchildren)

    hrm = ' '.join(['-not -path ' + repr(pat) for pat in exclude_patterns])
    if hrm:
        hrm = ' ' + hrm
    command = f"""{_find_command} -not -path '.operations*'{hrm}"""
    # TODO failover to builtin rglob
    return _transitive_(path, command)


def transitive_dirs(path):
    """Fast list of all child directories using unix find."""
    if sys.platform == 'win32':  # no findutils
        gen = os.walk(path)
        next(gen)  # drop path itself to avoid drp == Path('.')
        return [path / t[0] for t in gen]

    command = f"""{_find_command} -type d"""
    # TODO failover to builtin rglob + filter
    return _transitive_(path, command)


def unicode_truncate(string, length):
    """ Truncate unicode string to at most maximum length.
        If truncation splits a multibyte charachter then ignore
        the final truncated bytes. """
    return string.encode()[:length].decode(errors='ignore')


class ApiWrapper:
    """ Sometimes you just need one more level of indirection!
    Abstract base class to wrap Blackfynn and Pennsieve apis.
    """

    _id_class = None
    _api_class = None
    _sec_remote = None
    _dp_class = None

    @classmethod
    def _get_connection(cls, project_id):#, retry=10):
        try:
            return cls._api_class(
                api_token=auth.user_config.secrets(
                    cls._sec_remote, project_id, 'key'),
                api_secret=auth.user_config.secrets(
                    cls._sec_remote, project_id, 'secret'))
        except KeyError as e:
            msg = (f'need record in secrets for {cls._sec_remote} '
                   f'organization {project_id}')
            raise exc.MissingSecretError(msg) from e

        #except Exception as e:  # was absent dsn caching + rate limits
            #from time import sleep
            #if 0 < retry:
                # exponential falloff with the final wait being 10 seconds
                #sleep(10 ** (1 / retry))
                #cls._get_connection(project_id, retry=retry - 1)
            #else:
                #raise e

    def __init__(self, project_id, anchor=None):
        # no changing local storage prefix in the middle of things
        # if you want to do that create a new class

        if isinstance(project_id, self._id_class):
            # FIXME make the _id_class version the internal default
            project_id = project_id.id

        import requests  # SIGH
        self._requests = requests
        self.bf = self._get_connection(project_id)
        self.organization = self.bf.context
        self.project_name = self.bf.context.name
        self.root = self.organization.id
        self._project_id = project_id  # keep it around for set_state

    def __getstate__(self):
        state = self.__dict__
        if 'bf' in state:
            state.pop('bf')  # does not pickle well due to connection

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.bf = self._get_connection(self._project_id)

    def create_package(self, local_path):
        pkg = self._dp_class(local_path.name, package_type='Generic')  # TODO mimetype -> package_type ...
        pcache = local_path.parent.cache
        pkg.dataset = pcache.dataset.id
        if pcache.id != pkg.dataset:
            pkg.parent = pcache.id

        # FIXME this seems to create an empty package with no files?
        # have to do aws upload first or something?
        return self.bf._api.packages.create(pkg)

    def get(self, id, attempt=1, retry_limit=3):
        log.log(9, 'We have gone to the network!')  # too verbose for debug
        if isinstance(id, self._id_class):
            # FIXME inver this so the id form is internal or implement
            # __str__ differently from __repr__
            id = id.id

        if id.startswith('N:dataset:'):
            try:
                thing = self.bf.get_dataset(id)  # heterogenity is fun!
            except Exception as e:
                if 'No dataset matching name or ID' in str(e):
                    # sigh no error types
                    raise exc.NoRemoteFileWithThatIdError(id) from e
                else:
                    raise e

        elif id.startswith('N:organization:'):
            if id == self.root:
                return self.organization  # FIXME staleness?
            else:
                # if we start form local storage prefix for everything then
                # this would work
                msg = 'TODO org does not match need other api keys.'
                raise Exception(msg)
        else:
            try:
                thing = self.bf.get(id)
            except self._requests.exceptions.HTTPError as e:
                resp = e.response
                if resp.status_code == 404:
                    msg = (f'{resp.status_code} {resp.reason!r} '
                           f'when fetching {resp.url}')
                    raise exc.NoRemoteFileWithThatIdError(msg) from e

                log.exception(e)
                thing = None
            except self._remotebase.UnauthorizedException as e:
                log.error(f'Unauthorized to access {id}')
                thing = None

        if thing is None:
            if attempt > retry_limit:
                msg = f'No remote object retrieved for {id}'
                raise exc.NoMetadataRetrievedError(msg)
            else:
                thing = self.get(id, attempt + 1)

        return thing

    def get_file_url(self, id, file_id):
        resp = self.bf._api.session.get(
            f'{self.bf._api._host}/packages/{id}/files/{file_id}')
        if resp.ok:
            resp_json = resp.json()
        elif resp.status_code == 404:
            msg = (f'{resp.status_code} {resp.reason!r} '
                   f'when fetching {resp.url}')
            raise exc.NoRemoteFileWithThatIdError(msg)
        else:
            resp.raise_for_status()

        try:
            return resp_json['url']
        except KeyError as e:
            raise e


def make_bf_cache_as_classes(BaseNode, File, Collection, Dataset, Organization):

    class FakeBFLocal(ApiWrapper):
        class bf:
            """ tricksy hobbitses """
            @classmethod
            def get_dataset(cls, id):
                #import inspect
                #stack = inspect.stack(0)
                #breakpoint()
                #return CacheAsDataset(id)
                class derp:
                    """ yep, this is getting called way down inside added
                        in the extras pipeline :/ """
                    doi = None
                    status = 'FAKE STATUS ;_;'
                return derp

        def __init__(self, project_id, anchor):
            self.organization = CacheAsBFObject(anchor)  # heh
            self.project_name = anchor.name
            self.project_path = anchor.local


    class CacheAsBFObject(BaseNode):
        def __init__(self, cache):
            self.cache = cache
            self.id = cache.id
            self.cache.meta

        @property
        def name(self):
            return self.cache.name

        @property
        def parent(self):
            parent_cache = self.cache.parent #.local.parent.cache
            if parent_cache is not None:
                return self.__class__(parent_cache)

        @property
        def parents(self):
            parent = self.parent
            while parent:
                yield parent
                parent = parent.parent

        def __iter__(self):
            for c in self.cache.local.children:
                cache = c.cache
                if cache.is_dataset():
                    child = CacheAsDataset(cache)
                elif cache.is_organization():
                    child = CacheAsOrganization(cache)
                elif cache.is_dir():
                    child = CacheAsCollection(cache)
                elif cache.is_broken_symlink():
                    child = CacheAsFile(cache)

                yield child

        @property
        def members(self):
            return []


    class CacheAsFile(CacheAsBFObject, File): pass
    class CacheAsCollection(CacheAsBFObject, Collection): pass
    class CacheAsDataset(CacheAsBFObject, Dataset):
        """ yep, this is getting called way down inside added
            in the extras pipeline :/ """
        doi = None
        status = 'FAKE STATUS ;_;'
        @property
        def created_at(self): return self.cache.meta.created
        @property
        def updated_at(self): return self.cache.meta.updated
        @property
        def owner_id(self): return self.cache.meta.user_id
    class CacheAsOrganization(CacheAsBFObject, Organization): pass
    return (FakeBFLocal, CacheAsBFObject, CacheAsFile,
            CacheAsCollection, CacheAsDataset, CacheAsOrganization)


def make_bf_id_regex(top_level_domain):
    """ Generate the abstracted regex for a given
    blackfynn/pennsieve identifier scheme """

    uuid4_regex = ('([0-9a-f]{8}-'
                   '[0-9a-f]{4}-'
                   '4[0-9a-f]{3}-'
                   '[89ab][0-9a-f]{3}-'
                   '[0-9a-f]{12})')
    canonical_regex = (
        # XXX this is not implemented right now but could be
        f'^https://api.{top_level_domain}/id/N:([a-z]):' + uuid4_regex + '$'
    )
    curie_regex = '([a-z]+):' + uuid4_regex
    id_regex = 'N:' + curie_regex
    types = ('package', 'collection', 'dataset', 'organization', 'user', 'team')
    paths = 'viewer', 'files', *[t + 's' for t in types]  # WHO LOVES CLASS SCOPE WHEEEE LOL PYTHON
    path_elem_regex = '([a-z]+)'

    uri_api_regex = (f'https://api.{top_level_domain}/' + path_elem_regex + '/' + id_regex + '(?:/' + path_elem_regex + '/([^/]+))?/?')
    org_pref = f'https://app.{top_level_domain}/N:([a-z]+):'
    uri_human_regex = (org_pref + uuid4_regex + '/' + path_elem_regex +
                       '(?:/' + id_regex +
                       '(?:/' + path_elem_regex +
                       '(?:/(?:' + id_regex + '|' +
                       '([^/]+)' +
                       '(?:/' + id_regex + ')?))?' +
                       ')?)?/?'
                       )
    compiled = [
        (re.compile('^' + rx + '$'), rx_type) for rx, rx_type in
        # FIXME curie_regex likely needs to support /files/123 suffix as well?
        zip((curie_regex, id_regex, uuid4_regex, uri_api_regex, uri_human_regex),
            ('curie', 'id', 'uuid', 'uri_api', 'uri_human'),)]

    return (re.compile(uuid4_regex), canonical_regex, curie_regex, id_regex,
            types, paths, path_elem_regex, uri_api_regex, org_pref,
            uri_human_regex, compiled)


class BlackfynnId(idlib.Identifier):
    """ put all static information derivable from a blackfynn id here """

    top_level_domain = 'blackfynn.io'
    (uuid4_regex, canonical_regex, curie_regex, id_regex,
     types, paths, path_elem_regex, uri_api_regex, org_pref,
     uri_human_regex, compiled) = make_bf_id_regex(top_level_domain)

    def __reduce__(self):
        return self.__class__, (self.id, self.type, self.file_id)

    def __new__(cls, id_uri_curie_uuid, type=None, file_id=None):
        # TODO validate structure
        # TODO binary uuid
        if isinstance(id_uri_curie_uuid, cls):
            return id_uri_curie_uuid

        for rx, match_type in cls.compiled:
            match = rx.match(id_uri_curie_uuid)
            if match is not None:
                break
        else:
            msg = f'{id_uri_curie_uuid.encode()!r} matched no known pattern'
            raise idlib.exc.MalformedIdentifierError(msg)

        groups = [g for g in match.groups() if g is not None]

        if match_type == 'uuid':
            if type is None:
                raise TypeError('A raw uuid must be acompanied by a type.')

            id_type = type
            uuid, = groups
            # FIXME this one probably should be resolved since it is composed

        elif match_type == 'uri_api' and 'files' in groups:
            *_, id_type, uuid, _, file_id_string = groups
            file_id = int(file_id_string)
        elif match_type == 'uri_human':
            # strip trailing garbage that is not a uuid
            _mgs = groups
            while _mgs and not cls.compiled[2][0].match(_mgs[-1]):
                _mgs = _mgs[:-1]

            *_, id_type, uuid = _mgs
        else:
            # there is a bunch of garbage up front
            *_, id_type, uuid = groups

        if type is not None and type != id_type:
            raise TypeError(f'type mismatch! {type} != {id_type}')

        id = f'N:{id_type}:{uuid}'

        self = super().__new__(cls, id)
        self.id = id
        self.type = id_type
        self.uuid = uuid
        self.file_id = file_id  # XXX NOTE yes, we are putting a
        # file_id on everything because it simplifies all the checking
        # FIXME this conflicts with aug.PathMeta api where we will need
        # to unpack file_id because id and file_id are dissociated
        return self

    def __init__(self, *args, **kwargs):
        self._identifier = self.curie # XXXXXXXXXXXX FIXME this is probably going to cause issues

    def __str__(self):
        # uri_api would be a more consistent approach but there is a balance
        return self.curie # XXXXXXXXXXXX FIXME this is probably going to cause issues

    def __repr__(self):
        file_id = '' if self.file_id is None else f', file_id={self.file_id}'
        return f'{self.__class__.__name__}({self.id}{file_id})'

    def __hash__(self):
        return hash((self.__class__, self.id, self.file_id))

    def __eq__(self, other):
        return (type(self) == type(other) and
                self.id == other.id and
                self.file_id == other.file_id)  # works because None == None

    def __lt__(self, other):
        return (type(self) == type(other) and
                self.id < other.id)

    @property
    def label(self):
        return self.curie

    @property
    def curie(self):
        return self.type + ':' + self.uuid

    @property
    def uri_api(self):
        # NOTE: this cannot handle file ids
        if self.type == 'package':
            if self.file_id is None:
                endpoint = 'packages/' + self.id
            else:
                endpoint = f'packages/{self.id}/files/{self.file_id}'
        elif self.type == 'collection':
            endpoint = 'collections/' + self.id
        elif self.type == 'dataset':
            endpoint = 'datasets/' + self.id
        elif self.type == 'organization':
            endpoint = 'organizations/' + self.id
        elif self.type == 'user':
            endpoint = 'users/' + self.id
        else:
            raise NotImplementedError(f'TODO type {self.type} needs api endpoint')

        return f'https://api.{self.top_level_domain}/' + endpoint

    def uri_human(self, organization=None, dataset=None):
        if self.type == 'package':
            # FIXME file_id?
            endpoint = f'/{organization.id}/datasets/{dataset.id}/files/v/{self.id}'
        elif self.type == 'collection':
            endpoint = f'/{organization.id}/datasets/{dataset.id}/files/{self.id}'
        elif self.type == 'dataset':
            endpoint = f'/{organization.id}/datasets/{self.id}'
        elif self.type == 'organization':
            endpoint =         f'/{self.id}/datasets'
        else:
            raise NotImplementedError(f'{self.type} TODO need app endpoint')

        return f'https://app.{self.top_level_domain}' + endpoint

    def asStr(self):
        # FIXME conflict with the fact that these ids are default
        # local and unqualified, but use uuids which are unique
        #return self.uri_api
        return self.id


class PennsieveId(BlackfynnId):
    # FIXME this isn't really a subclass but you know how bad oo
    # implementations tend to conflate type with functionality ...

    top_level_domain = 'pennsieve.io'
    (uuid4_regex, canonical_regex, curie_regex, id_regex,
     types, paths, path_elem_regex, uri_api_regex, org_pref,
     uri_human_regex, compiled) = make_bf_id_regex(top_level_domain)


class LocId(idlib.Identifier):

    def __init__(self, id, type):
        self.id = id
        self.type = type
        self.curie = id  # FIXME depends on how id was generated
        sysid, path = id.split(':', 1)
        if '/' not in path:
            raise TypeError(f'WHAT HAVE YOU DONE {path}')
        self.uri_api = 'file://' + path
        self._identifier = self.uri_api

    def uri_human(self, *args, **kwargs):
        return self.uri_api



import os
import io
import json
from copy import deepcopy
from future.utils import string_types
import requests
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from .config import auth
from sparcur.utils import log


def bind_agent_command(agent_module):
    # FIXME this is a nearly perfect use case for orthauth wrapping
    default_agent_cmd = agent_module.agent_cmd
    def agent_cmd():
        rcp = auth.get_path('remote-cli-path')
        if rcp is None:
            return default_agent_cmd()
        else:
            return rcp.as_posix()

    agent_module.agent_cmd = agent_cmd


@property
def patch_session(self):
    """
    Make requests-futures work within threaded/distributed environment.
    """
    if self._session is None:
        self._session = Session()
        self._set_auth(self._token)
        try:
            backoff_factor = auth.get('remote-backoff-factor')
        except Exception as e:
            log.exception(e)
            backoff_factor = 1

        # Enable retries via urllib
        adapter = HTTPAdapter(
            pool_connections=1000,  # wheeee
            pool_maxsize=1000,  # wheeee
            max_retries=Retry(
                total=self.settings.max_request_timeout_retries,
                backoff_factor=backoff_factor,
                status_forcelist=[502, 503, 504] # Retriable errors (but not POSTs)
            )
        )
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)

    return self._session


def Blackfynn_get(self, id, update=True):
    return self._api.core.get(id, update=update)


def File_download(self, destination):  # FIXME not clear that we use this?
    """ remove prefix functionality since there are filenames without
        extensions ... """

    if self.type == "DirectoryViewerData":
        raise NotImplementedError(
            "Downloading S3 directories is currently not supported")

    if os.path.isdir(destination):
        # destination dir
        f_local = os.path.join(destination, os.path.basename(self.s3_key))
    else:
        # exact location
        f_local = destination

    r = requests.get(self.url, stream=True)
    with io.open(f_local, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)

    # set local path
    self.local_path = f_local

    return f_local


def PackagesAPI_get(self, pkg, include='files,source'):
    """
    Get package object

    pkg:     can be DataPackage ID or DataPackage object.
    include: list of fields to force-include in response (if available)
    """
    pkg_id = self._get_id(pkg)

    params = {'includeAncestors': 'true'}
    if include is not None:
        if isinstance(include, string_types):
            params.update({'include': include})
        elif hasattr(include, '__iter__'):
            params.update({'include': ','.join(include)})

    resp = self._get(self._uri('/{id}', id=pkg_id), params=params)

    # TODO: cast to specific DataPackages based on `type`
    pkg = self._get_package_from_data(resp)
    pkg._resp = resp
    return pkg


# package meta


def bind_packages_File(File):
    class FakeBFile(File):
        """ Fake file to simplify working with package metadata """

        id = None  # unforunately these don't seem to have made it through

        def __init__(self, package, **kwargs):
            self.package = package
            self._json = kwargs
            def move(*tuples):
                for f, t in tuples:
                    kwargs[t] = kwargs.pop(f)

            move(('createdAt', 'created_at'),
                ('updatedAt', 'updated_at'),
                ('fileType', 'type'),
                ('packageId', 'pkg_id'),
            )

            if 'size' not in kwargs:
                kwargs['size'] = None  # if we have None on a package we know it is not zero

            if 'checksum' in kwargs:
                cs = kwargs['checksum']
                kwargs['chunksize'] = cs['chunkSize']
                kwargs['checksum'] = cs['checksum']  # overwrites but ok
            else:
                kwargs['checksum'] = None

            for k, v in kwargs.items():
                setattr(self, k, v)

        @property
        def state(self):
            return self.package.state

        @property
        def owner_id(self):
            return self.package.owner_id

        @property
        def dataset(self):
            return self.package.dataset

        @property
        def parent(self):
            # cut out the middle man (hopefully)
            return self.package.parent

        def __repr__(self):
            return ('<' +
                    self.__class__.__name__ +
                    ' '.join(f'{k}={v}' for k, v in self._json.items()) +
                    '>')


    def _packages(self,
                  pageSize=1000,
                  includeSourceFiles=True,
                  raw=False,
                  latest_only=False,
                  filename=None):
        """ python implementation to make use of /dataset/{id}/packages """
        remapids = {}

        def restructure(j):
            """ restructure package json to match what api needs? """
            # FIXME something's still wonky here
            c = j['content']
            c['int_id'] = c['id']
            c['id'] = c['nodeId']  # FIXME packages seem to be missing ids!?
            remapids[c['int_id']] = c['id']
            c['int_datasetId'] = c['datasetId']
            c['datasetId'] = c['datasetNodeId']
            if 'parentId' in c:
                pid = c['parentId']
                c['parent'] = remapids[pid]  # key error to signal out of order
                #if pid in remapids:
                #else:
                    #c['parent'] = f'WTF ERROR: {pid}'
                    #print('wtf', pid, c['id'], c['datasetId'])
                #else:
                    #c['parent'] = remapids['latest']
            return j

        index = {self.id: self}  # make sure that dataset is in the index
        session = self._api.session
        # cursor
        # pageSize
        # includeSourceFiles
        # types
        # filename
        filename_args = f'&filename={filename}' if filename is not None else ''
        cursor_args = ''
        out_of_order = []
        while True:
            try:
                resp = session.get(
                    f'{self._api._host}/datasets/{self.id}/packages?'
                    f'pageSize={pageSize}&'
                    f'includeSourceFiles={str(includeSourceFiles).lower()}'
                    f'{filename_args}'
                    f'{cursor_args}')
            except requests.exceptions.RetryError as e:
                log.exception(e)
                # sporadic 504 errors that we probably need to sleep on
                breakpoint()
                raise e

            #print(resp.url)
            if resp.ok:
                j = resp.json()
                packages = j['packages']
                if raw:
                    yield from packages
                    if latest_only:
                        break
                    else:
                        continue

                if out_of_order:
                    packages += out_of_order
                    # if a parent is on the other side of a pagination boundary put
                    # the children at the end and move on
                out_of_order = [None]
                while out_of_order:
                    #log.debug(f'{out_of_order}')
                    if out_of_order[0] is None:
                        out_of_order.remove(None)
                    elif packages == out_of_order:
                        if filename is not None:
                            out_of_order = None
                        elif 'cursor' not in j:
                            raise RuntimeError('We are going nowhere!')
                        else:
                            # the missing parent is in another castle!
                            break
                    else:
                        packages = out_of_order
                        out_of_order = []
                    for count, package in enumerate(packages):
                        if isinstance(package, dict):
                            id = package['content']['nodeId']
                            name = package['content']['name']
                            bftype = self._id_to_type(id)
                            dcp = deepcopy(package)
                            try:
                                rdp = restructure(dcp)
                            except KeyError as e:
                                if out_of_order is None:  # filename case
                                    # parents will simply not be listed
                                    # if you are using filename then beware
                                    if 'parentId' in dcp['content']:
                                        dcp['content'].pop('parentId')
                                    rdp = restructure(dcp)
                                else:
                                    out_of_order.append(package)
                                    continue

                            bfobject = bftype.from_dict(rdp, api=self._api)
                            if name != bfobject.name:
                                log.critical(f'{name} != {bfobject.name}')
                            bfobject._json = package
                            bfobject.dataset = index[bfobject.dataset]
                        else:
                            bfobject = package

                        if (
                                isinstance(bfobject.parent, str)
                                and bfobject.parent in index):
                            parent = index[bfobject.parent]
                            if parent._items is None:
                                parent._items = []
                            parent.items.append(bfobject)
                            bfobject.parent = parent
                            # only put objects in the index when they have a
                            # parent that is a bfobject, this ensures that you
                            # can always recurse to base once you get an object
                            # from this function
                            index[bfobject.id] = bfobject
                            if parent.state == 'DELETING':
                                if not bfobject.state == 'DELETING':
                                    bfobject.state = 'PARENT-DELETING'
                            elif parent.state == 'PARENT-DELETING':
                                if not bfobject.state == 'DELETING':
                                    bfobject.state = 'PARENT-DELETING'

                            yield bfobject  # only yield if we can get a parent
                        elif out_of_order is None:  # filename case
                            yield bfobject
                        elif bfobject.parent is None:
                            # both collections and packages can be at the top
                            # level dataset was set to its bfobject repr above
                            # so safe to yield
                            if bfobject.dataset is None:
                                log.debug('No parent no dataset\n' +
                                          json.dumps(bfobject._json, indent=2))
                            index[bfobject.id] = bfobject
                            yield bfobject
                        else:
                            out_of_order.append(bfobject)
                            continue

                        if isinstance(bfobject, self._dp_class):
                            bfobject.fake_files = []
                            if 'objects' not in bfobject._json:
                                log.error(f'{bfobject} has no files!??!')
                            else:
                                for i, source in enumerate(
                                        bfobject._json['objects']['source']):
                                    # TODO package id?
                                    if len(source) > 1:
                                        log.info('more than one key in source '
                                                 f'{sorted(source)}')

                                    ff = FakeBFile(
                                        bfobject, **source['content'])
                                    bfobject.fake_files.append(ff)
                                    yield ff

                                    if i == 1:  # only log once
                                        msg = ('MORE THAN ONE FILE IN PACKAGE '
                                               f'{bfobject.id}')
                                        log.critical(msg)

                if 'cursor' in j:
                    cursor = j['cursor']
                    cursor_args = f'&cursor={cursor}'
                else:
                    break

            else:
                break

    return FakeBFile, _packages


@property
def packages(self, pageSize=1000, includeSourceFiles=True):
    yield from self._packages(
        pageSize=pageSize,
        includeSourceFiles=includeSourceFiles)


def packagesByName(self,
                   pageSize=1000,
                   includeSourceFiles=True,
                   filenames=tuple()):
    if filenames:
        for filename in filenames:
            yield from self._packages(
                pageSize=pageSize,
                includeSourceFiles=includeSourceFiles,
                filename=filename)
    else:
        yield from self._packages(
            pageSize=pageSize,
            includeSourceFiles=includeSourceFiles)


@property
def packageTypeCounts(self):
    session = self._api.session
    resp = session.get(
        f'{self._api._host}/datasets/{self.id}/packageTypeCounts')
    if resp.ok:
        j = resp.json()
        return j
    else:
        resp.raise_for_status()


@property
def publishedMetadata(self):
    session = self._api.session
    resp = session.get(
        f'{self._api._host}/discover/search/datasets?query={self.int_id}')
    if resp.ok:
        j = resp.json()
        if j['totalCount'] == 1:
            return j['datasets'][0]
        elif j['totalCount'] == 0:
            return
        else:
            cands = [d for d in j['datasets'] if d['sourceDatasetId'] == self.int_id]
            lc = len(cands)
            if lc == 1:
                return cands[0]
            elif lc == 0:
                return
            else:
                return sorted(cands, key=lambda d:d['version'], reverse=True)[0]

        return j
    else:
        resp.raise_for_status()


def publishedVersionMetadata(self, id_published, published_version):
    session = self._api.session
    resp = session.get((
        f'{self._api._host}/discover/datasets/'
        f'{id_published}/versions/{published_version}/metadata'))
    if resp.ok:
        return resp.json()
    else:
        resp.raise_for_status()


@property
def Dataset_users(self):
    session = self._api.session
    resp = session.get(
        f'{self._api._host}/datasets/{self.id}/collaborators/users')
    return resp.json()


@property
def Dataset_teams(self):
    session = self._api.session
    resp = session.get(
        f'{self._api._host}/datasets/{self.id}/collaborators/teams')
    return resp.json()


@property
def Dataset_doi(self):
    session = self._api.session
    resp = session.get(
        f'{self._api._host}/datasets/{self.id}/doi')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code != 404:
            log.warning(f'{self} doi {resp.status_code}')


def Dataset_delete(self):
    """ actually delete """
    session = self._api.session
    resp = session.delete(f'{self._api._host}/datasets/{self.id}')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code != 404:
            log.warning(f'{self} issue deleting {resp.status_code}')


@property
def Dataset_contributors(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/datasets/{self.id}/contributors')
    return resp.json()


@property
def Dataset_banner(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/datasets/{self.id}/banner')
    return resp.json()


@property
def Dataset_readme(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/datasets/{self.id}/readme')
    return resp.json()


@property
def Dataset_status_log(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/datasets/{self.id}/status-log')
    return resp.json()


@property
def Dataset_meta(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/datasets/{self.id}')
    return resp.json()


@property
def Organization_teams(self):
    session = self._api.session
    resp = session.get(f'{self._api._host}/organizations/{self.id}/teams')
    return resp.json()



{'config-search-paths': ['{:user-config-path}/sparcur/config.yaml',],
 'auth-variables':
 {'data-path': {
     'default': None,
     'environment-variables': 'SPARCUR_DATA_PATH SPARC_DATA_PATH DATA_PATH'},
  'export-path': {
     'default': '{:user-data-path}/sparcur/export',
     'environment-variables':
     'SPARCUR_EXPORT_PATH SPARC_EXPORTS EXPORT_PATH'},
  'cache-path': {
      'default': '{:user-cache-path}/sparcur',
      'environment-variables': 'SPARCUR_CACHE_PATH CACHE_PATH'},
  'cleaned-path': {
     'default': '{:user-data-path}/sparcur/cleaned',
     'environment-variables':
     'SPARCUR_CLEANED_PATH SPARC_CLEANED CLEANED_PATH'},
  'log-path': {
      'default': '{:user-log-path}/sparcur',
      'environment-variables': 'SPARCUR_LOG_PATH LOG_PATH'},
  'resources': {
      'default': [
          '../resources/',  # git
          '{:cwd}/share/sparcur/resources',  # ebuild testing
          '{:user-data-path}/sparcur/resources',  # pip install --user
          '{:prefix}/share/sparcur/resources',  # system
          '/usr/share/sparcur/resources',],  # pypy3
      'environment-variables': 'SPARCUR_RESOURCES'},
  'export-url': {
      'default': None,
      'environment-variables': 'SPARCUR_EXPORT_URL'},
  'remote-cli-path': {
      'default': None,
      'environment-variables': 'REMOTE_CLI_PATH'},
  'remote-organization': {  # FIXME cryptic error if this is not set
      # idlib.exceptions.MalformedIdentifierError: b'None' matched no known pattern
      'environment-variables':
      'BLACKFYNN_ORGANIZATION PENNSIEVE_ORGANIZATION REMOTE_ORGANIZATION'},
  'remote-backoff-factor': {
      'default': 1,
      'environment-variables': 'BLACKFYNN_BACKOFF_FACTOR'},
  'google-api-service-account-file-readonly': None,
  'google-api-service-account-file-rw': None,
  'hypothesis-group-name': {'environment-variables': 'HYP_GROUP_NAME'},
  'hypothesis-api-key': {'environment-variables': 'HYP_API_KEY HYP_API_TOKEN'},
  'hypothesis-group': {'environment-variables': 'HYP_GROUP'},
  'hypothesis-user': {'environment-variables': 'HYP_USER'},
  'preview': {
      'default': False,
      'environment-variables': 'SPARCUR_PREVIEW'},
  'never-update': False,
  'datasets-noexport': None,
  'datasets-sparse': None,
  'datasets-no': None,
  'datasets-test': None,
  'sparse-limit': {
      'default': 10000,
      'environment-variables': 'SPARCUR_SPARSE_LIMIT SPARSE_LIMIT'},}}



import os
if 'PENNSIEVE_LOG_LEVEL' not in os.environ:
    # silence agent import warning
    os.environ['PENNSIEVE_LOG_LEVEL'] = 'CRITICAL'
from pennsieve import log as _pnlog
# blackfynn.log sets logging.basicConfig which pollutes logs from
# other programs that are sane and do not use the root logger
# so we have to undo the damage done by basic config here
# we add the sparcur local handlers back in later
from sparcur.utils import log, silence_loggers
for __pnlog in (_pnlog.get_logger(), _pnlog.get_logger("pennsieve.agent")):
    silence_loggers(__pnlog)
    __pnlog.addHandler(log.handlers[0])

from pennsieve import Pennsieve, DataPackage, BaseNode
from pennsieve import Organization, Dataset, Collection, File
from pennsieve import base as pnb
from pennsieve.api import agent
from pennsieve.api.data import PackagesAPI
from sparcur import monkey
from sparcur.utils import ApiWrapper, PennsieveId, make_bf_cache_as_classes


def id_to_type(id):
    #if isinstance(id, BlackfynnId):  # FIXME this is a bad place to do this (sigh)
        #return {'package': DataPackage,
                #'collection':Collection,
                #'dataset': Dataset,
                #'organization': Organization,}[id.type]

    if id.startswith('N:package:'):
        return DataPackage
    elif id.startswith('N:collection:'):
        return Collection
    elif id.startswith('N:dataset:'):
        return Dataset
    elif id.startswith('N:organization:'):
        return Organization


class PNLocal(ApiWrapper):

    _id_class = PennsieveId
    _api_class = Pennsieve
    _sec_remote = 'pennsieve'
    _dp_class = DataPackage
    _remotebase = pnb


monkey.bind_agent_command(agent)

FakeBFile, _packages = monkey.bind_packages_File(File)

# monkey patches

Dataset._dp_class = DataPackage
Dataset.delete = monkey.Dataset_delete
Dataset.meta = monkey.Dataset_meta
Dataset.packagesByName = monkey.packagesByName
Dataset.packageTypeCounts = monkey.packageTypeCounts
Dataset.publishedMetadata = monkey.publishedMetadata
Dataset.publishedVersionMetadata = monkey.publishedVersionMetadata
Dataset.readme = monkey.Dataset_readme
Dataset.contributors = monkey.Dataset_contributors
Dataset.doi = monkey.Dataset_doi
Dataset.status_log = monkey.Dataset_status_log  # XXX NOTE this overwrites a method
Dataset.packages = monkey.packages
Dataset._packages = _packages
Pennsieve.get = monkey.Blackfynn_get
#PackagesAPI.get = monkey.PackagesAPI_get


(FakeBFLocal, CacheAsBFObject, CacheAsFile,
 CacheAsCollection, CacheAsDataset, CacheAsOrganization
 ) = make_bf_cache_as_classes(BaseNode, File, Collection, Dataset, Organization)



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::utils.py][utils.py]]
from sparcur.config import auth
__doc__ = f"""Common command line options for all sparcur.simple modules
Usage:
    sparcur-simple manifest  [options]  [<path>...]
    sparcur-simple get-uuid  <remote-id>...
    sparcur-simple datasets
    sparcur-simple for-racket
    sparcur-simple check-names
    sparcur-simple git-repos [update] [options]
    sparcur-simple [options] [<path>...]
    sparcur-simple [options] [--dataset-id=<ID>...]
    sparcur-simple [options] [--extension=<EXT>...] [<path>...]

Commands:
    manifest                        generate manifest files for path
    for-racket                      print data for reading into Racket

Options:
    -h --help                       show this

    --hypothesis-group-name=NAME    the hypotheis group name

    --project-id=ID                 the project id
    --dataset-id=<ID>...            one or more datset ids
    --project-id-auth-var=VAR       name of the auth variable holding the project-id

    --project-path=PATH             the project path, will be path if <path>... is empty
    --parent-path=PATH              the parent path where the project will be cloned to
    --parent-parent-path=PATH       parent in which a random tempdir is generated
                                    or the dataset uuid is used as the parent path, don't use this ...
    --invariant-local-path=PATH     path relative to parent path for dataset [default: dataset]
    --export-local                  set export-parent-path to {{:parent-path}}/exports/
    --export-path=PATH              base export path containing the standard path structure [default: {auth.get_path('export-path')}]
    --cleaned-parent-path=PATH      direct parent path into which cleaned paths will be placed
    --cleaned-path=PATH             base cleaned path containing the standard path structure [default: {auth.get_path('cleaned-path')}]
    --export-parent-path=PATH       direct parent path into which exports will be placed
    --extension=<EXT>...            one or more file extensions to fetch

    -j --jobs=N                     number joblib jobs [default: 12]
    --exclude-uploaded              do not pull files from remote marked as uploaded
    --sparse-limit=N                package count that forces a sparse pull [default: {auth.get('sparse-limit')}]
    --symlink-objects-to=PATH       path to an existing objects directory
    --log-level=LEVEL               log level info [default: INFO]
    --open=PROGRAM                  show file with PROGRAM
    --show                          show file with xopen
    --pretend                       show what would be done for update
"""

import os
import sys
from types import GeneratorType
from pyontutils import clifun as clif
from sparcur import exceptions as exc
from sparcur.utils import _find_command
from sparcur.utils import log, logd, loge, GetTimeNow, PennsieveId
from sparcur.paths import Path, PennsieveCache
from sparcur.backends import PennsieveRemote


def backend_pennsieve(project_id=None, Local=Path, Cache=PennsieveCache):  # (ref:def_babf)
    """return a configured pennsieve backend
        calling this is sufficient to get everything set up correclty

        You must call RemotePath.init(project_id) before using
        RemotePath.  Passing the project_id argument to this function
        will do that for you. It is not required because there are
        cases where the project_id may not be known at the time that
        this function is called. """

    RemotePath = PennsieveRemote._new(Local, Cache)

    if project_id is not None:
        RemotePath.init(project_id)

    return RemotePath


class Options(clif.Options):

    @property
    def id(self):
        # dataset_id has priority since project_id can occure without a
        # dataset_id, but dataset_id may sometimes come with a project_id
        # in which case the dataset_id needs priority for functions that
        # branch on the most granular identifier type provided
        return (self.dataset_id[0]
                if self.dataset_id else
                (self.project_id
                 if self.project_id else
                 None))

    @property
    def project_id(self):
        if not hasattr(self, '_cache_project_id'):
            id = self._args['--project-id']
            if id is not None:
                identifier = PennsieveId(id, type='organization')
                self._cache_project_id = identifier
            else:
                return

        return self._cache_project_id

    @property
    def dataset_id(self):
        if not hasattr(self, '_cache_dataset_id'):
            ids = self._args['--dataset-id']
            if ids:
                identifiers = [PennsieveId(id, type='dataset') for id in ids]
                self._cache_dataset_id = identifiers
            else:
                return ids

        return self._cache_dataset_id

    @property
    def remote_id(self):
        if not hasattr(self, '_cache_remote_id'):
            ids = self._args['<remote-id>']
            if ids:
                identifiers = [PennsieveId(id) for id in ids]
                self._cache_remote_id = identifiers
            else:
                return ids

        return self._cache_remote_id

    @property
    def jobs(self):
        return int(self._args['--jobs'])

    n_jobs = jobs  # match internal kwargs conventions

    @property
    def paths(self):
        return [Path(p).expanduser().resolve() for p in self._args['<path>']]

    @property
    def path(self):
        paths = self.paths
        if paths:
            return paths[0]
        elif self.project_path:
            return self.project_path
        else:
            # if no paths were listed default to cwd
            # consistent with how the default kwargs
            # are set on a number of mains
            # this is preferable to allow path=None
            # to be overwritten by the conventions of
            # individual pipeline mains
            return Path.cwd()

    @property
    def project_path(self):
        pp = self._args['--project-path']
        if pp:
            return Path(pp).expanduser().resolve()

    @property
    def parent_parent_path(self):
        ppp = self._args['--parent-parent-path']
        if ppp:
            return Path(ppp).expanduser().resolve()
        else:
            return Path.cwd()

    @property
    def parent_path(self):
        pap = self._args['--parent-path']
        did = self.dataset_id
        if pap:
            return Path(pap).expanduser().resolve()
        elif did:
            id = self.id  # dataset_id is a list so use id which handles that
            uuid = id.uuid
            return (self.parent_parent_path / uuid).expanduser().resolve()

    @property
    def export_path(self):
        ep = self._args['--export-path']
        epp = self.export_parent_path
        if ep and epp:
            raise TypeError('Only one of --export-path and --export-parent-path may be set.')
        elif ep:
            return Path(ep).expanduser().resolve()
        else:
            raise Exception('should not happen')

    @property
    def export_parent_path(self):
        epp = self._args['--export-parent-path']
        el = self.export_local
        pap = self.parent_path
        if epp and el:
            raise TypeError('Only one of --export-local and --export-parent-path may be set.')
        elif epp:
            return Path(epp).expanduser().resolve()
        elif el and pap:
            # it is ok to do this here becuase the TypeError above prevents
            # the case where both epp and el are set, so even though epp
            # is no longer always what was set on the command line, it is
            # it is the case that there won't be conflicting sources
            return pap / 'exports'

    @property
    def cleaned_path(self):
        cp = self._args['--cleaned-path']
        cpp = self.export_parent_path
        if cp and cpp:
            raise TypeError('Only one of --cleaned-path and --cleaned-parent-path may be set.')
        elif cp:
            return Path(cp).expanduser().resolve()
        else:
            raise Exception('should not happen')

    @property
    def cleaned_parent_path(self):
        cpp = self._args['--cleaned-parent-path']
        if cpp:
            return Path(cpp).expanduser().resolve()

    @property
    def extensions(self):
        return self.extension

    @property
    def symlink_objects_to(self):
        slot = self._args['--symlink-objects-to']
        if slot:
            return Path(slot).expanduser()

    @property
    def sparse_limit(self):  # FIXME not being pulled in by asKwargs ??
        return int(self._args['--sparse-limit'])

    @property
    def time_now(self):  # FIXME make it possible to pass this in?
        if not hasattr(self, '_time_now'):
            self._time_now = GetTimeNow()

        return self._time_now

    @property
    def log_level(self):  # FIXME not being pulled in by asKwargs ??
        ll = self._args['--log-level']
        if ll.isdigit() or ll[0] == '-' and ll[1:].isdigit():
            return int(ll)
        else:
            return ll


def pipe_main(main, after=None, argv=None):
    options, args, defaults = Options.setup(__doc__, argv=argv)
    # _entry_point is used as a way to switch behavior when a
    # pipeline main is run directly or actually in a pipeline
    try:
        log.setLevel(options.log_level)
        logd.setLevel(options.log_level)
        loge.setLevel(options.log_level)
        out = main(_entry_point=True, **options.asKwargs())
    except Exception as e:
        log.exception(e)
        log.error(options.path)
        raise e

    if after:
        after(out)

    return out


def rglob(path, pattern):
    """ Hack around the absurd slowness of python's rglob """

    if sys.platform == 'win32':
        log.warning('no findutils on windows, watch out for unexpected files')
        return list(path.rglob(pattern))

    doig = (hasattr(path, 'cache') and
            path.cache and
            path.cache.cache_ignore)
    exclude = ' '.join([f"-not -path './{p}*'" for p in path.cache.cache_ignore]) if doig else ''
    command = f"""{_find_command} {exclude} -name {pattern!r}"""

    with path:
        with os.popen(command) as p:
            string = p.read()

    path_strings = string.split('\n')  # XXX posix path names can contain newlines
    paths = [path / s for s in path_strings if s]
    return paths


def _fetch(cache):  # sigh joblib multiprocessing pickling
    # lambda functions are great right up until you have to handle an
    # error function inside of them ... thanks python for yet another
    # failure to be homogenous >_<
    meta = cache.meta
    try:
        size_mb = meta.size.mb
    except AttributeError as e:
        if meta.errors:
            logd.debug(f'remote errors {meta.errors} for {cache!r}')
            return
        else:
            raise exc.StopTheWorld(cache) from e

    return cache.fetch(size_limit_mb=size_mb + 1)  # FIXME somehow this is not working !?


def _fetch_path(path):  # sigh joblib multiprocessing pickling
    path = Path(path)
    cache = path.cache
    if cache is None:
        raise exc.NoCachedMetadataError(path)

    # do not return to avoid cost of serialization back to the control process
    _fetch(cache)


def fetch_paths_parallel(paths, n_jobs=12, use_async=True):
    if n_jobs <= 1:
        [_fetch_path(path) for path in paths]
    elif use_async:
        from pyontutils.utils import Async, deferred
        Async()(deferred(_fetch_path)(path) for path in paths)
    else:
        import pathlib
        from joblib import Parallel, delayed
        backend = 'multiprocessing' if hasattr(sys, 'pypy_version_info') else None
        # FIXME FIXME FIXME somehow this can result in samples.xlsx being fetched to subjects.xlsx !?!??!!
        # XXX either a race condition on our end or something insane from the remote
        Parallel(n_jobs=n_jobs, backend=backend)(delayed(_fetch_path)(pathlib.Path(path)) for path in paths)
        #Parallel(n_jobs=n_jobs)(delayed(fetch_path)(path) for path in paths)


def combinate(*functions):
    def combinator(*args, **kwargs):
        for f in functions:
            # NOTE no error handling is needed here
            # in no cases should the construction of
            # the python object version of a path fail
            # all failures should happen _after_ construction
            # the way we have implemented this they fail when
            # the data attribute is accessed
            obj = f(*args, **kwargs)
            if isinstance(obj, GeneratorType):
                yield from obj
                # FIXME last one wins, vs yield tuple vs ...?
                # FIXME manifests are completely broken for this
            else:
                yield obj

    return combinator


def multiple(func, merge=None):
    """ combine multiple results """
    def express(*args, **kwargs):
        vs = tuple(func(*args, **kwargs))
        if merge is not None:
            yield merge(vs)
        else:
            yield vs

    return express


def early_failure(path, error, dataset_blob=None):
    # these are the 9999 5555 and 4444 errors
    # TODO match the minimal schema reqs as
    # we did in pipelines
    if dataset_blob is None:
        cache = path.cache
        return {'id': cache.id,
                'meta': {'uri_api': cache.uri_api,
                         'uri_human': cache.uri_human,},
                #'status': 'early_failure',  # XXX note that status is not requried
                # if we cannot compute it, but if there are inputs there should be
                # a status
                'errors': [error],  # FIXME format errro
                }

    else:
        if 'errors' not in datset_blob:
            dataset_blob['errors'] = []

        datset_blob['errors'].append(error)
        return dataset_blob


class DataWrapper:
    # sigh patterns are stupid, move this to elsewhere so it doesn't taint everything
    def __init__(self, data):
        self.data = data


def main(id=None, **kwargs):
    def ik(key):
        return key in kwargs and kwargs[key]

    if id is not None:
        print(id.uuid)

    if ik('get_uuid'):
        for id in kwargs['remote_id']:
            print(id.uuid)

        return

    if (ik('datasets') or ik('for_racket') or ik('check_names')):
        log.setLevel(60)  # logging.CRITICAL = 50
        from sparcur.config import auth
        from sparcur.simple.utils import backend_pennsieve
        if ik('project_id'):
            pass  # project_id from command line
        else:
            project_id = auth.get('remote-organization')

        PennsieveRemote = backend_pennsieve(project_id)
        root = PennsieveRemote(project_id)
        datasets = list(root.children)

    if ik('datasets'):
        print('\n'.join([d.id for d in datasets]))

    if ik('for_racket'):
        import json
        _dsmeta = '\n'.join([f"({json.dumps(d.id)} {json.dumps(d.name)})"
                             for d in datasets])
        dsmeta = f"({_dsmeta})"
        # lab pi last name should be cached in some other way
        print(dsmeta)

    if ik('check_names'):
        # you will want to run sparcur.simple.fetch_remote_metadata_all
        from pathlib import PurePosixPath
        def report(pid, exp, pub):
            pname = pub['name']
            name_mismatch = (
                False if exp['basename'] == pname
                else (exp['basename'], pname))
            # [6:] to strip files/
            ppname = PurePosixPath(pub['path']).name
            pathname_mismatch = (
                False if exp['basename'] == ppname
                else (exp['basename'], ppname))

            eppp = PurePosixPath(exp['dataset_relative_path']).parent.as_posix()
            pppp = PurePosixPath(pub['path'][6:]).parent.as_posix()
            parent_path_mismatch = (
                False if eppp == pppp
                else (eppp, pppp))

            # once we fix things on our end names should match
            # parent paths should match
            # name and pathname might match but do not have to match
            pid = f'  {pid}'
            nsp = '\n        '
            if name_mismatch:
                if pathname_mismatch and pname != ppname:
                    return (f'{pid}    name mismatch and pathname mismatch                '
                            f'{nsp}{nsp.join(name_mismatch)}{nsp}{nsp.join(pathname_mismatch)}')
                else:
                    return (f'{pid}    name mismatch                                      '
                            f'{nsp}{nsp.join(name_mismatch)}')

            if parent_path_mismatch:
                if True:
                    return (f'{pid}                                        parent mismatch'
                            f'{nsp}{nsp.join(parent_path_mismatch)}')

            if True:
                if True:
                    return ''
                    #return (f'{pid} ok                                                    ')

        import json
        from sparcur.backends import PennsieveDatasetData
        export_path = kwargs['export_path']
        epd = export_path / 'datasets'
        for dataset in datasets:
            latest = epd / dataset.identifier.uuid / 'LATEST'
            export_path_metadata =  latest / 'path-metadata.json'
            exported = export_path_metadata.exists()

            # pass the remote not just the id so that bfobject is
            # accessible to the RemoteDatasetData class
            pdd = PennsieveDatasetData(dataset)
            rmeta = pdd.fromCache()
            published = 'id_published' in rmeta
            if exported and published:
                with open(export_path_metadata, 'rt') as f:
                    j = json.load(f)
                epni = {'N:' + o['remote_id']:o for o in j['data']
                        if o['remote_id'].startswith('package:')}
                ppni = pdd._published_package_name_index()
                se, sp = set(epni), set(ppni)
                e_missing = sp - se
                p_missing = se - sp
                s_common = sp & se
                rep = [report(c, epni[c], ppni[c]) for c in s_common]
                repstr = '\n'.join([r for r in rep if r])
                if repstr:
                    print(dataset.id, 'bad')
                    print(repstr)
                else:
                    print(dataset.id, 'ok')
            elif exported:
                print(dataset.id, 'unpublished')
            elif published:
                print(dataset.id, 'unexported')
            else:
                print(dataset.id, 'unpublished and unexported')

    if ik('git_repos'):
        import augpathlib as aug
        import sparcur
        from sparcur.config import auth
        from importlib import metadata
        d = metadata.distribution(sparcur.__package__)
        rps = [p for p in [aug.RepoPath(d._path) for d in d.discover()] if p.working_dir]
        setups = [p for p in [p.parent / 'setup.py' for p in rps] if p.exists()]
        wds = sorted(set([p.working_dir for p in rps]))
        never_update = auth.get('never-update')
        pretend=ik('pretend') or never_update
        if pretend:
            if never_update:
                print(f'never-update: true is set in {auth.user_config._path}')
            print('These are the commands that would be run.')
        def doupdate(rp):
            if pretend:
                print(f'git -C {rp.as_posix()} stash')
                print(f'git -C {rp.as_posix()} pull')
                return
            print(f'Pulling {rp.as_posix()}')
            print(rp.repo.git.stash())
            # TODO checkout to a safety branch and tag for rollback
            print(rp.repo.git.pull())
        if ik('update'):
            for wd in wds:
                doupdate(wd)
            # indescriminately run setup.py with --release set to tangle
            from importlib import util as imu
            oldargv = sys.argv
            try:
                for s in setups:
                    if pretend:
                        print(f'pushd {s.parent.as_posix()}; python setup.py --release; popd')
                        continue
                    sys.argv = ['setup.py', '--release']  # reset every time since most will mutate
                    print(f'Maybe tangling via {s.as_posix()}')
                    spec = imu.spec_from_file_location(f'setup_{s.parent.name}', s)
                    mod = imu.module_from_spec(spec)
                    try:
                        with s.parent:  # ah relative paths
                            spec.loader.exec_module(mod)
                    except SystemExit:
                        pass
                    except Exception as e:
                        log.exception(e)
            finally:
                sys.argv = oldargv

    if ik('manifest'):
        from sparcur.utils import write_manifests
        paths = kwargs['paths']
        if not paths:
            paths = [Path.cwd()]

        manifests_rendered = write_manifests(parents=paths)
        manifests, rendered = zip(*manifests_rendered)
        nl = '\n'
        print(f'generated manifests at:\n{nl.join([m.as_posix() for m in manifests])}')
        if ik('open'):
            cmd = kwargs['open']
            [m.xopen(cmd) for m in manifests]
        elif ik('show'):
            [m.xopen() for m in manifests]


if __name__ == '__main__':
    pipe_main(main)
# utils.py ends here



# term mapping

from functools import wraps
from .core import OntTerm
from .utils import log


def tos(f):
    @wraps(f)
    def inner(v):
        if isinstance(v, str):
            return f(v)
        elif isinstance(v, tuple):
            return tuple(f(_) for _ in v)
        elif isinstance(v, list):
            return [f(_) for _ in v]

    return inner


# TODO load from db/config ?

_species = {
    'canis lupus familiaris': OntTerm('NCBITaxon:9615',  label='Canis familiaris'),
    'felis catus':            OntTerm('NCBITaxon:9685',  label='Felis catus'),
    'guinea pig':             OntTerm('NCBITaxon:10141', label='Cavia porcellus'),
    'homo sapiens':           OntTerm('NCBITaxon:9606',  label='Homo sapiens'),
    'mus musculus':           OntTerm('NCBITaxon:10090', label='Mus musculus'),
    'mustela putorius furo':  OntTerm('NCBITaxon:9669',  label='Mustela putorius furo'),
    'rattus norvegicus':      OntTerm('NCBITaxon:10116', label='Rattus norvegicus'),
    'suncus murinus':         OntTerm('NCBITaxon:9378',  label='Suncus murinus'),
    'sus scrofa':             OntTerm('NCBITaxon:9823',  label='Sus scrofa'),
    'sus scrofa domesticus':  OntTerm('NCBITaxon:9825',  label='Sus scrofa domesticus'),
}


@tos
def species(string, __species=dict(_species), __fetched=[False]):
    #if not __fetched[0]:  # SIGH
        #[v.fetch() for v in __species.values()]  # TODO parallel
        #__fetched[0] = True

    lstr = string.lower()
    if lstr in __species:
        return __species[lstr]
    else:
        log.warning(f'No ontology mapping found for {string}')
        return string


_sex = {
    'female': OntTerm('PATO:0000383', label='female'),
    'male': OntTerm('PATO:0000384', label='male'),
}


@tos
def sex(string, __sex=dict(_sex), __fetched=[False]):
    #if not __fetched[0]:  # SIGH
        #[v.fetch() for v in __sex.values()]  # TODO parallel
        #__fetched[0] = True

    lstr = string.lower()
    if lstr in __sex:
        return __sex[lstr]
    else:
        log.warning(f'No ontology mapping found for {string}')
        return string



#!/usr/bin/env python3

from types import MappingProxyType
import yaml
import idlib
from github import Github
from sparcur import exceptions as exc
from sparcur.config import auth
from sparcur.core import adops
from sparcur.utils import log
from sparcur.export import latest_ir
install_requires = ['PyGithub']


def condense(thing, delist=True):
    if isinstance(thing, dict):
        return {k:nv for k, nv in [(k, condense(v)) for k, v in thing.items()]
                if nv}
    elif isinstance(thing, list):
        out = [_ for _ in [condense(t) for t in thing] if _]
        if delist and len(out) == 1:
            return out[0]
            
        return out
    else:
        return thing


def _issubclass(thing, cls):
    """ LOL PYTHON """
    # why would you raise a type error instead of returning false !??!?!
    try:
        return issubclass(thing, cls)
    except TypeError:
        return False


class DictObject:
    """ static accessor for dictionary objects

        operators
        / any-key
        * over-list -> [['further', 'path', 'elements']]
    """

    _mapping = MappingProxyType({})

    # TODO consider generating these from the json schema or mapping the classes via
    # the schema ... the convention I use here is a bit of a pain to create

    def __new__(cls, blob):
        cls = type(cls.__name__ + '_DictObject', (cls,), {})
        for attr, path in cls._mapping.items():  # FIXME consider doing path -> attr instead so the could be combined
            term_list = path[-1] if isinstance(path[-1], list) else None
            term = (path[-1]
                    if _issubclass(path[-1], DictObject)
                    else ((path[-1][0] if
                           _issubclass(path[-1][0], DictObject) else
                           None) if
                          term_list else
                          None))

            #log.debug(term)
            if term_list or term:
                path = path[:-1]

            #@property
            #@idlib.utils.cache_result  # FIXME this does NOT work because the function name is always f
            def f(self, path=path, terminal=term, term_list=term_list):
                try:
                    end = adops.get(self._blob, path)  # FIXME / and * would need to be implemented on top of this
                except exc.NoSourcePathError:
                    if term_list:
                        return []
                    elif terminal is not None:
                        return terminal({})
                    else:
                        return []  # FIXME was return None but that breaks missing keys that should contain lists

                # FIXME type check
                if terminal and term_list:
                    return [terminal(o) for o in end]
                elif terminal:
                    return terminal(end)
                elif term_list:
                    raise NotImplementedError('hrm')
                else:
                    return end


            setattr(f, '__name__', attr)

            setattr(cls, attr, property(idlib.utils.cache_result(f)))

        return object.__new__(cls)

    def __init__(self, blob):
        self._blob = blob


class Meta(DictObject):

    _mapping = {
        'uri_human': ['uri_human'],
        'uri_api': ['uri_api'],
        '_award_number': ['award_number'],
        'folder_name': ['folder_name'],
        # FIXME need to figure out how to preserve the type :/
        'protocols': ['protocol_url_or_doi'],  # FIXME merge multiple paths ? or collect from subjects to meta as well?
        'created': ['timestamp_created'],
    }

    @property
    @idlib.utils.cache_result
    def award_number(self):
        an  = self._award_number
        if an:
            return Award(an)


class Contributor(DictObject):

    _mapping = {
        'id': ['id'],
        'first_name': ['first_name'],
        'last_name': ['last_name'],
        'orcid': ['contributor_orcid_id'],
        'role': ['contributor_role'],
        'contact': ['is_contact_person'],
        }


class Dataset(DictObject):

    # FIXME allowing the mapping to go only a single path deep
    # would mean that we wouldn't have to explicitly create
    # a mapping for all the keys, but could inspect the blob to generate
    # any that were not specified
    # XXX ON THE OTHER HAND: if you start using d.meta.value and d.meta is None
    # you are in a world of pain :/ SIGH
    _mapping = {
        'id': ['id'],
        'meta': ['meta', Meta],
        'contributors': ['contributors', [Contributor]],  # FIXME probably need to spec the internal type
        #'uri_human': ['meta', 'uri_human'],
        #'uri_api': ['meta', 'uri_api'],
        #'award_number': ['meta', 'award_number'],
        #'folder_name': ['meta', 'folder_name'],
        #'protocols': ['meta', 'protocol_url_or_doi'],  # FIXME merge multiple paths ? or collect from subjects to meta as well?
    }

    def asIssue(self, repo):
        return DatasetIssue(self, repo)


class OtAwardId(idlib.Identifier, idlib.Stream):

    _local_conventions = idlib.conventions.Identity

    def __init__(self, identifier):
        self._identifier = identifier  # FIXME if this is a list or something other than a string, __str__ breaks below

    def __str__(self):
        id = self.identifier
        if not isinstance(id, str):
            raise TypeError(id)

        return id

    def __repr__(self):
        return self.__class__.__name__ + f"({self.identifier!r})"


class Award(idlib.Stream):
    # FIXME TODO and here we see why we want to have resolvers as their own thing
    # multiple award ids would make use of a single stream type
    # in this case it is the "scicrunch reporter metadata stream"
    # and there can and will be many more than one _id_class that could
    # be associated with this stream/resolver, the question is how
    # to handle that non-1:1-ness
    _id_class = OtAwardId  # FIXME ... _id_classes ? and then set _id_class @ instance level?
    _resolver_template = ('https://scicrunch.org/scicrunch/data/source/'
                          'nif-0000-10319-1/search?q={id}')
    identifier_actionable = idlib.StreamUri.identifier_actionable
    dereference_chain = idlib.StreamUri.dereference_chain
    dereference = idlib.StreamUri.dereference
    headers = idlib.StreamUri.headers

    def __init__(self, award_number):
        self._identifier = self._id_class(award_number)

    def scicrunchUri(self, asType=None):
        uri_string = self._resolver_template.format(id=self.identifier)
        return uri_string if asType is None else asType(uri_string)


class GithubRepo(idlib.Stream):
    # substreams
    # git repo
    # wiki git repo
    # issues
    # other api stuff

    # XXX GithubRepo isn't really a Stream because it is not
    # in principle immutable in the way we want/need streams to be
    # things can change rapidly and thus synchronization of state
    # has to happen much more frequently than with streams ...
    # maybe idlib.Dynamic ? or idlib.Storm ? or idlib.River ?
    # idlib.Well idlib.Pool idlib.Creek idlib.Sewer idlib.Whirlpool
    # idlib.Sea idlib.Cloud (lol) idlib.Rain idlib.Tsunami ...
    # need more dynamic water terms ... idlib.Flood

    def __init__(self, fragment):
        self._identifier = fragment
        # FIXME orthauth this and defer setup
        github = Github(auth.user_config.secrets('github',
                                                 'tgbugs-build',
                                                 'WARNING-development-all-scopes'))
        #repo = github.get_repo('SciCrunch/sparc-datasets')
        self.repo = github.get_repo(self._identifier)
        self._issues = []

    @property
    def issues(self):
        if not self._issues:
            self._pull_issues()

        return self._issues

    def _pull_issues(self):
        self._issues = self.repo.get_issues()


class GithubIssue(idlib.Stream):
    # TODO set tags
    pass


class IssueMapper:
    """ Base class for mapping external ids to their corresponding
        issue tracker identifier. """


class DatasetIssue(IssueMapper):
    """ class that maps datasets to their github issue """

    def __init__(self, dataset, repo):
        # TODO path vs blob vs id + latest
        self.dataset = dataset
        self.repo = repo  # should be a GithubRepo not a github.Repo
        existing = self.exists()
        if existing:  # FIXME catch multiple issue
            self._issue = existing[0]

    @property
    def dataset_id(self):
        return self.dataset.id

    @property
    def issue_id(self):
        # TODO uri_human = self._issue.html_url
        return self._issue.number

    @property
    def new_issue_title(self):
        return f'Tracking {self.dataset_id}'

    def parse_top_comment(self):
        """ needed for updates """

    @staticmethod
    def _markdown_link(href, text=None):
        if text is not None:
            return f'[{text}]({href})'
        else:
            return href

    @property
    def dataset_link(self):
        try:
            return self._markdown_link(self.dataset.meta.uri_human, self.dataset.id)
        except AttributeError as e:
            breakpoint()
            raise e

    @property
    def dataset_award(self):
        if self.dataset.meta.award_number:
            id = self.dataset.meta.award_number.identifier.identifier
            if id:
                return (f'{id} '
                        f'{self._markdown_link(self.dataset.meta.award_number.scicrunchUri(), "SciCrunch")}')

    @property
    def dataset_protocols(self):
        return [self._markdown_link(p.identifier, p.label_safe) for p in self.dataset.meta.protocols]

    @property
    def dataset_doi(self):
        if hasattr(self.dataset.meta, 'doi') and self.dataset.meta.doi:
            doi = self.dataset.meta.doi
            return f'{doi}'

    def new_issue_blob(self):
        a, b = self.new_issue_blobs()
        return {**a, **b}

    def _contrib_fmt(self, c, flag=True):
        bad = '' if 'orcid.org' in c.id or not flag else '*'
        # FIXME for some reason c.id are strings after loading ir?
        return self._markdown_link(c.id, f'{c.first_name} {c.last_name}{bad}')

    def new_issue_blobs(self):
        pis = [c for c in self.dataset.contributors
               if 'PrincipalInvestigator' in c.role]
        blob_a = {
            'Dataset': self.dataset_link,
            'Title': self.dataset.meta.folder_name,
            'Award': self.dataset_award,
            'Lab': [p.last_name for p in pis][:1],  # FIXME last name also lab might not be the first PI last name
            'People': {
                'PI': [self._contrib_fmt(p) for p in pis],
                'Contact': [self._contrib_fmt(c) for c in self.dataset.contributors
                            if ('Contact' in c.role or c.contact) and c not in pis],
            },
            'Protocols': self.dataset_protocols,
            'Created': self.dataset.meta.created,
            'Published': self.dataset_doi,
        }

        blob_b = {
            'isAbout': self.dataset.meta.uri_api,  # FIXME TODO how to embed/rediscover source ids
            # TODO consider embedding the github issue id in the rmeta somehow/somewhere
        }
        return condense(blob_a), condense(blob_b)

    def new_issue_comment(self):
        # FIXME justification
        # NOTE we retain the quotes around hyperlinks to simplify roundtrip
        blobs = self.new_issue_blobs()
        one, two = [yaml.dump(_, default_flow_style=False, sort_keys=False)
                    for _ in blobs]
        return one + '\n----------------\n' + two

    @property
    def new_issue_labels(self):
        return []

    def exists(self):
        title = self.new_issue_title
        self.repo._pull_issues
        return [i for i in self.repo.issues if title == i.title]

    def issue_update(self):
        """update issue from dataset"""
        body = self.new_issue_comment()
        self._issue.edit(body=body)

    def new_issue_create(self):
        if self.exists():
            raise ValueError('Issue already exists! {self.uri_human}')
        #resp = self.submit_new_issue(data)
        title = self.new_issue_title
        body = self.new_issue_comment()
        labels = self.new_issue_labels
        issue = self.repo.repo.create_issue(title=title,
                                            body=body,
                                            labels=labels)
        self._issue = issue
        return issue

    def embed_id_tracker(self):
        """ embed the tracker id in the source metadata somehow """
        if self.id_tracker is not None:
            self.stream


class ProtocolIssue(IssueMapper):
    """ TODO """
    # facet link
    # etc.


def main():
    #print(comment)
    import orthauth as oa

    github = Github(auth.user_config.secrets('github',
                                             'tgbugs-build',
                                             'WARNING-development-all-scopes'))

    grepo = github.get_repo('SciCrunch/sparc-datasets')

    commits = list(grepo.get_commits())
    blob = oa.utils.QuietDict(latest_ir())
    ds = [Dataset(d) for d in blob['datasets']]
    [print(DatasetIssue(d).new_issue_comment()) for d in ds]
    breakpoint()


if __name__ == '__main__':
    main()



import unittest
from sparcur.core import adops, DictTransformer
from sparcur.derives import Derives as De


class Examples:
    # yield expect, data, args
    @property
    def add(self):
        yield {'hello': 'world'}, {}, (['hello'], 'world')
        yield {'hello': {'there': {'general': 'world'}}}, {}, (['hello', 'there', 'general'], 'world')
        yield {'hello': {'there': {'general': 'world'}}}, {'hello': {}}, (['hello', 'there', 'general'], 'world')
        yield {'world': [{'hello': 'there'},
                         {'hello': 'there'}]}, {'world': [{},
                                                          {}]}, (['world', int, 'hello'], 'there')
        yield {'world': {'oh': 'no'}}, {'world': {'oh': 'no'}}, (['world', 'hello', int, 'mr'], 'there')

    @property
    def add_error(self):
        yield {'hello':'1'}, {'hello':'1'}, (['hello'], 'there')
        yield {'hello':'2'}, {'hello':'2'}, ([], 'there')
        yield {'hello':'3'}, {'hello':'3'}, ('lolol', 'there')  # this might succeed :x indeed it did
        yield {'hello': {'there': {'general': 'world'}}}, {'hello': 'OH NO'}, (['hello', 'there', 'general'], 'world')

    @property
    def get(self):
        yield {'hello': 'world'}, {'hello','world'}, (['hello'],)
        yield {'hello': {'there': {'general': 'world'}}}, {}, (['hello', 'there', 'general'],)

    @property
    def get_error(self):
        yield {'hello': 'world'}, {'hello','world'}, (['another'],)

    @property
    def delete(self):
        yield {}, {'hello','world'}, (['hello'],)
        yield ({'hello': {'there': {'keep': 'MEEE'}}},
               {'hello': {'there': {'general': 'world', 'keep': 'MEEE'}}},
               (['hello', 'there', 'general'],))

    @property
    def delete_error(self):
        yield {'hello': 'world'}, {'hello','world'}, (['hello', 'world'],)

    @property
    def copy(self):
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['hello'], ['another'])

    @property
    def copy_error(self):
        yield {'another':'world', 'hello': 'world'}, {'another':'world', 'hello': 'world'}, (['hello'], ['another'])
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, ([], ['another'])
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['hello'], [])  # dicts are closed world can't break on through
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['no-source'], ['oops'])

    @property
    def move(self):
        yield {'another':'world'}, {'hello': 'world'}, (['hello'], ['another'])
        yield {'source': {},
               'target': {'b': 1}}, {'source': {'a': 1},
                                     'target': {}}, (['source', 'a'], ['target', 'b'])

    @property
    def move_error(self):
        yield {'another':'1'}, {}, (['hello'], ['another'])
        yield {'another':'2'}, {'another': 0, 'hello':'world'}, (['hello'], ['another'])

    @property
    def update(self):
        yield [0, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6], ([0], 0)

    @property
    def update_error(self):
        # tuple object does not support item assignment
        yield (0, 2, 3, 4, 5, 6), (1, 2, 3, 4, 5, 6), ([0], 0)


class TestDer(unittest.TestCase):
    def test_der(self):
        assert De.contributor_name('b, a') == ('a', 'b')


class ExamplesDT:

    @property
    def move(self):
        # after before command
        sko = dict(source_key_optional=True)

        # simple moves
        yield ({'other-key': 'something'},
               {'other-key': 'something'},
               [[['not-here'], ['not-here', 'over-there']]],
               sko)

        yield ({'not-here': {'over-there': {}}},
               {'not-here': {}},
               [[['not-here'], ['not-here', 'over-there']]],
               sko)

        yield ({'not-here': {'over-there': []}},
               {'not-here': []},
               [[['not-here'], ['not-here', 'over-there']]],
               sko)

        yield ({'not-here': {'over-there': None}},
               {'not-here': None},
               [[['not-here'], ['not-here', 'over-there']]],
               sko)

        # pattern matching with holes
        yield ({'a': [{'c': 1}, {'c': 2}]},

               {'a': [{'b': 1}, {'b': 2}]},

               [[['a', int, 'b'],
                 ['a', int, 'c']]],
               sko)

        yield ({'a': {'x': [{'c': 1}, {'c': 2}]}},

               {'a': {'x': [{'b': 1}, {'b': 2}]}},

               [[['a', 'x', int, 'b'],
                 ['a', 'x', int, 'c']]],
               sko)

        return  # things not implemented yet
        yield (
            {'d': [{'g': [{'h': 1}, {'h': 2}]},
                   {'g': [{'h': 3}, {'h': 4}]}]},

            {'d': [{'e': [{'f': 1}, {'f': 2}]},
                   {'e': [{'f': 3}, {'f': 4}]}]},

            [[['d', int, 'e', int, 'f'],
              ['d', int, 'g', int, 'h']]],

            sko)

        # TODO it is not clear whether the examples below are moves

        yield ({'d': [{'e': [{}, {}],
                       'k': [1, 2]},
                      {'e': [{}, {}],
                       'k': [3, 4]}]},
               {'d': [{'e': [{'f': 1}, {'f': 2}]},
                      {'e': [{'f': 3}, {'f': 4}]}]},
               [[['d', int, 'e', int, 'f'],
                 ['d', int, 'k', int]]],
               sko)

        yield ({'z': [{'e': [{}, {}],
                       'g': [1, 2]},
                      {'e': [{}, {}],
                       'g': [3, 4]}]},
               {'d': [{'e': [{'f': 1}, {'f': 2}]},
                      {'e': [{'f': 3}, {'f': 4}]}]},
               [[['d', int, 'e', int, 'f'],
                 ['z', int, 'g', int]]],
               sko)

        # note that moving individual elements of a list i.e. ending
        # the source-path with int does not make sense because that
        # would change the structure of the source along the way

    @property
    def derive(self):
        yield ({'a': 2, 'b': 3, 'c': 5, 'd': 6, 'e': 1},
               {'a': 2, 'b': 3}, [[[['a'], ['b']],
                                   lambda a, b: (a + b, a * b),
                                   [['c'], ['d']]],
                                  # look you can chain them!
                                  [[['c'], ['d']],
                                   lambda c, d: (d - c,),
                                   [['e']]]])
        yield ({'name': 'b, a', 'first_name': 'a', 'last_name': 'b'},
               {'name': 'b, a'}, [[[['name']],
                                   De.contributor_name,
                                   [['first_name'], ['last_name']]]])

        # this is no longer an error because we allow TypeError and
        # catch the issue using the schema
        yield ({'a': 1}, {'a': 1}, [[[['a']],
                                     # TODO is there any way to catch this information early ?
                                     lambda _: None,  # previously None would include a TypeError
                                     [['b']]]])

    @property
    def lift(self):
        yield {'hello':'there'}, {'hello': 'world'}, [[['hello'], lambda v: 'there']]

    @property
    def update(self):
        yield {'hello':[
            'there',
            'there',]}, {'hello': [
                'general',
                'kenobi',
            ]}, [[['hello', int], lambda v: 'there']]

        yield {'hello':[
            {'n': 1},
            {'n': 1}]}, {'hello': [
                {'n': 0},
                {'n': 0}]}, [[['hello', int, 'n'], lambda v: v + 1]]


def failed_to_error(function, data, args, kwargs):
    try:
        function(data, *args, **kwargs)
        print(function, data, args)
        return True
    except BaseException as e:
        return False


class Populator:
    apply = True
    @classmethod
    def populate(cls):
        for name in cls.functions:
            name_error = name + '_error'
            name_negative = name + '_negative'

            test_name = 'test_' + name
            function_to_test = getattr(cls.to_test, name)
            def fun(self, name=name, function=function_to_test):
                bads = [(expect, data, args)
                        for expect, data, args, *kwargs in getattr(self, name)
                        if not function(data,
                                        *(args if cls.apply else (args,)),
                                        **{k:v for kw in kwargs for k, v in kw.items()})
                        and data != expect]
                assert not bads, bads

            setattr(cls, test_name, fun)

            if hasattr(cls, name_error):
                test_name_error = 'test_' + name_error
                def efun(self, name=name_error, function=function_to_test):
                    bads = [(expect, data, args)
                            for expect, data, args, *kwargs in getattr(self, name)
                            if failed_to_error(function,
                                               data,
                                               (args if cls.apply else (args,)),
                                               {k:v for kw in kwargs for k, v in kw.items()})]
                    assert not bads, bads

                setattr(cls, test_name_error, efun)

            if hasattr(cls, name_negative):
                test_name_negative = 'test_' + name_negative
                def nfun(self, name=name_negative, function=function_to_test):
                    bads = [(expect, data, args)
                            for expect, data, args, *kwargs in getattr(self, name)
                            if not function(data,
                                            *(args if cls.apply else (args,)),
                                            **{k:v for kw in kwargs for k, v in kw.items()})
                            and data == expect]
                    assert not bads, bads

                setattr(cls, test_name_negative, nfun)


class TestAdops(Examples, Populator, unittest.TestCase):
    functions = 'add', 'copy', 'move', 'update'
    to_test = adops
TestAdops.populate()


class TestDictTransformer(ExamplesDT, Populator, unittest.TestCase):
    functions = 'lift', 'derive', 'update', 'move',
    to_test = DictTransformer
    apply = False
TestDictTransformer.populate()



import os
import json
import unittest
import pytest
from pyontutils.utils import Async, deferred
from sparcur.core import JEncode
from sparcur.extract import xml as exml
from .common import examples_root, RDHPN

export = False


class TestExtractMetadata(unittest.TestCase):

    def test_new_mbf_format(self):
        x = examples_root / 'mbf-example.xml'
        embf = exml.ExtractXml(x)
        d = embf.asDict()
        errors = d.pop('errors') if 'errors' in d else tuple()
        error_types = set(e['validator'] for es in errors for e in es)
        assert error_types == {'not'} or not error_types, f'unexpected error type! {error_types}'


class ExtractMetadataReal:

    def test_mbf_header(self):
        test_id = 'N:dataset:bec4d335-9377-4863-9017-ecd01170f354'
        test_dataset = [d.cache for d in self.test_datasets if d.cache.id == test_id][0]
        if not list(test_dataset.local.children):
            rchilds = list(test_dataset.rchildren)
            xmls = [c for c in rchilds if c.suffix == '.xml']
            Async(rate=5)(deferred(x.fetch)() for x in xmls if not x.exists())
            #[x.fetch() for x in xmls if not x.exists()]
            local_xmls = [x.local for x in xmls]
        else:
            local_xmls = list(test_dataset.local.rglob('*.xml'))
            if any(p for p in local_xmls if not p.exists()):
                raise BaseException('unfetched children')

        embfs = [exml.ExtractXml(x) for x in local_xmls]
        d = embfs[0].asDict()
        blob = [e.asDict() for e in embfs]
        errors = [b.pop('errors') for b in blob if 'errors' in b]
        error_types = set(e['validator'] for es in errors for e in es)
        if export:
            with open('mbf-test.json', 'wt') as f:
                json.dump(blob, f, indent=2, cls=JEncode)
            with open('mbf-errors.json', 'wt') as f:
                json.dump(errors, f, indent=2, cls=JEncode)

        assert error_types == {'not'} or not error_types, f'unexpected error type! {error_types}'


class TestExtractMetadataRealPN(RDHPN, ExtractMetadataReal, unittest.TestCase):
    pass



import copy
from typing import Tuple
from functools import wraps
from collections import defaultdict
import idlib
from sparcur import schemas as sc
from sparcur import exceptions as exc
from sparcur import normalization as nml
from sparcur.core import log, logd, JPointer, HasErrors
from sparcur.utils import register_type


def collect(*oops, unpacked=True):
    def decorator(generator_function):
        @wraps(generator_function)
        def inner(*args, **kwargs):
            out = tuple(generator_function(*args, **kwargs))
            if unpacked:
                return out
            else:
                return out,

        return inner

    if oops:
        generator_function, = oops
        return decorator(generator_function)
    else:
        return decorator


class Derives:
    """ all derives must return as a tuple, even if they are length 1 """

    @staticmethod
    def contributor_name(name) -> Tuple[str, str]: 
        if ', ' in name:
            last, first = name.split(', ', 1)
        elif ',' in name:
            last, first = name.split(',', 1)  # the error will be recorded by the json schema
        elif ' ' in name:
            first, last = name.split(' ', 1)
        else:
            first = 'ERROR IN NAME see lastName for content'
            last = name

        return first, last

    @staticmethod
    @collect
    def creators(contributors):
        for c in contributors:
            if 'contributor_role' in c and 'Creator' in c['contributor_role']:
                # FIXME diry diry mutation here that should happen in a documente way
                cont = {**c}  # instead of importing copy since this is a one deep
                cont.pop('contributor_role', None)
                cont.pop('is_contact_person', None)
                yield cont

    @staticmethod
    def award_number(raw_award_number, funding) -> str:
        return nml.NormAward(nml.NormAward(raw_award_number))

    @staticmethod
    def pi(contributors):
        pointers = [JPointer(f'/contributors/{i}')
                    for i, c in enumerate(contributors)
                    if (('PrincipalInvestigator' in c['contributor_role'])
                        if 'contributor_role' in c else False)]

        if len(pointers) == 1:
            pointer, = pointers
            return pointer

        return pointers

    @staticmethod
    def dataset_species(subjects) -> Tuple[tuple]:
        out = set()
        for subject in subjects:
            if 'species' in subject:
                out.add(subject['species'])

        if len(out) == 1:
            return next(iter(out))

        return tuple(out)

    @staticmethod
    def doi(doi_string):  # FIXME massive network sandbox violation here
        """ check if a doi string resolves, if it does, return it """
        if doi_string is None:
            raise TypeError('WHAT HAVE YOU DONE!?')

        doi = idlib.Doi(doi_string)
        try:
            metadata = doi.metadata()  # FIXME network sandbox violation
            if metadata is not None:
                return doi
        except idlib.exceptions.RemoteError:
            # sometimes a doi is present on the platform but does not resolve
            # in which case we don't add it as metadata because it has not
            # been officially published, just reserved, this check is more
            # correct than checkin the status on the platform
            # FIXME HOWEVER it violates the network sandbox, so we probably
            # need an extra step during the data retrieval phase which attempts
            # to fetch all the doi metadata
            pass
        except Exception as e:
            # XXX random errors need to be ignored here for now
            # since this really should not be run at this step
            # due to the network dependency, we need a post-network
            # step where we can strip out all the things that fail
            log.exception(e)

    @staticmethod
    def _lift_mr(path_dataset, dataset_relative_path, record, should_log):
        parent = dataset_relative_path.parent
        if 'filename' not in record:
            msg = f'filename missing from record in {dataset_relative_path}'
            raise exc.BadDataError(msg)

        record_drp = parent / record['filename']

        # FIXME this is validate move elsewhere ??? but this is also
        # where we want to embed data about the path ...
        # but I supposed we could inject errors later ? this is an
        # extremely obscure place to inject these ...
        # and to do that we need access to the path
        # and it is clear given that we passed in THIS_PATH

        # FIXME TODO how to deal with sparse datasets
        _record_path = path_dataset / record_drp  # do not include in export

        _cache = _record_path.cache
        if _cache is None:
            lifted = _record_path._jsonMetadata()  # will produce the error for us
        else:
            lifted = _cache._jsonMetadata()

        if 'errors' in lifted:
            he = HasErrors(pipeline_stage='Derives._lift_mr')
            _message = lifted['errors'].pop()['message']
            # FIXME pretty sure that path in addError is used for both
            # the json path and the file system path
            message = ('Non-existent path listed in manifest '
                       f'{dataset_relative_path}\n{_message}')
            if he.addError(message,
                           blame='submission',
                           path=_record_path):
                if should_log:
                    should_log = False
                    logd.error(message)

            he.embedErrors(lifted)

        lifted['prov:wasDerivedFrom'] = (
            # have to reattach path_dataset because Path.cwd() may not be
            # the dataset root (usually the organizaiton root)
            path_dataset / dataset_relative_path).cache_identifier
        lifted['dataset_relative_path'] = record_drp
        lifted['manifest_record'] = {
            k:v for k, v in record.items() if k != 'filetype'
        }
        if 'additional_types' in record:
            # FIXME TODO make sure that the mimetypes match
            # FIXME currently a string, either change name or make a list?
            lifted['mimetype'] = record['additional_types']

        return lifted, should_log

    @staticmethod
    def _scaffolds(lifted, scaffolds):
        # FIXME this is burried here, should be more like how we
        # handle the xml extraction

        expected = 'organ', 'species'
        if ('mimetype' in lifted and
            lifted['mimetype'] in ('inode/vnd.abi.scaffold+directory',
                                   'inode/x.vnd.abi.scaffold+directory',)):
            # TODO type -> scaffold
            # TODO look for additional metadata from interior manifest
            scaf = copy.deepcopy(lifted)
            # FIXME should be a move step for organ and specie
            for key in expected:
                if key in scaf['manifest_record']:
                    scaf[key] = scaf['manifest_record'][key]

            scaffolds.append(scaf)

    @classmethod
    def path_metadata(cls, path_dataset, manifests, xmls):
        path_metadata = []
        scaffolds = []  # FIXME need a better abstraction for additional known types e.g. the mbf segmentations
        for manifest in manifests:
            he = HasErrors(pipeline_stage='Derives.path_metadata')
            if 'contents' in manifest:
                contents = manifest['contents']
                if 'manifest_records' in contents:
                    drp = manifest['dataset_relative_path']
                    _should_log = True
                    for record in contents['manifest_records']:
                        try:
                            lifted, _should_log = cls._lift_mr(
                                path_dataset, drp, record, _should_log)
                        except FileNotFoundError as e:
                            # FIXME all this is horribly tangled and we repeatedly compute the same value
                            record_drp = drp.parent / record['filename']
                            msg = f'A path listed in manifest {drp} does not exist! {record_drp}'
                            if he.addError(msg, blame='submission', path=drp):
                                logd.error(e)
                            continue  # FIXME need this in the errors record
                        except exc.BadDataError as e:
                            if he.addError(e, blame='submission', path=drp):
                                logd.error(e)
                            continue  # FIXME need this in the errors record
                        path_metadata.append(lifted)
                        if 'errors' not in lifted:
                            cls._scaffolds(lifted, scaffolds)

                    he.embedErrors(manifest)

        path_metadata.extend(xmls['xml'])
        # TODO try to construct/resolve the referent paths in these datasets as well
        # getting the path json metadata will embed errors about missing files for us
        return path_metadata, scaffolds

    @staticmethod
    def samples_to_subjects(samples, schema_version=None):
        """ extract subject information from samples sheet """
        # TODO only allow this for <= 1.2.3 ? or no, really
        # only the subject id should be present in later versions

        # FIXME TODO derive subject fields from schema?
        subject_fields = (
            'subject_id',
            'ear_tag_number',
            'rrid_for_strain',
            'genus',
            'species',
            'strain',
            'sex',
            'gender',
            'age',
            'age_category',
            'body_weight',
            'body_mass',
            'body_mass_weight',
            'height_inches',
            'age_years',
            'organism_rrid',
            'age_range_min',
            'age_range_max',
            'date_of_euthanasia',
        )
        subjects = [{k:v for k, v in blob_sample.items() if k in subject_fields}
                    for blob_sample in samples]
        return subjects,

    @staticmethod
    def validate_structure(path, dir_structure, performances, subjects, samples):
        he = HasErrors(pipeline_stage='Derives.validate_structure')

        # FIXME TODO handle pools as well and figure out cases where subjects/samples are metadata only

        # for dataset templates of the 1.* series
        # general approach: set of all specimen ids and set of all
        # folder names take the ones that match ignore the known ok
        # that do not, and warn on all the rest that do not match
        # and that are not inside a known specimen or subject folder

        valid_top_123 = ('source', 'primary', 'derivative',  # FIXME not here :/ schema somehow?
                         'code', 'docs', 'protocol')

        def top_level(drp):
            return drp.parent.name == '' and drp.name in valid_top_123
            
        # absolute_paths = [path / pblob['dataset_relative_path'] for pblob in dir_structure]
        dd = defaultdict(list)
        for pblob in dir_structure:
            drp = pblob['dataset_relative_path']
            p = drp.parts
            dd[p[-1]].append((len(p), drp, p[::-1]))

        dirs = {k:av for k, vs in dd.items()
                for av in ([v for v in vs if not top_level(v[1])],)
                # cull empty in a single step
                if av} 

        perfs = {p['performance_id']:p for p in performances}

        # subject_id could be missing, but we filter failures on all of
        # those so in theory we shouldn't need to handle it as this stage
        subs = {s['subject_id']:s for s in subjects}

        dd = defaultdict(list)
        for s in samples:
            dd[s['sample_id']].append(s)
        samps = dict(dd)

        ### pools
        dd = defaultdict(list)
        for s, d in subs.items():
            if 'pool_id' in d:
                dd[d['pool_id']].append(s)
        sub_pools = dict(dd)
        pool_subs = {v:k for k, vs in sub_pools.items() for v in vs}

        dd = defaultdict(list)
        for s, d in samps.items():
            if d and 'pool_id' in d[0]:
                dd[d[0]['pool_id']].append(s)
        sam_pools = dict(dd)
        # XXX we can check for multiple pools per sample at various
        # stages including here, but technically a single specimen
        # being in multiple pools definitely happens
        pool_sams = {v:k for k, vs in sam_pools.items() for v in vs}

        records = []
        done_dirs = set()
        done_specs = set()

        ### performances
        union_perf = set(dirs) | set(perfs)
        inter_perf = set(dirs) & set(perfs)
        done_dirs.update(inter_perf)
        not_done_perfs = set(perfs) - inter_perf

        ### subject pools
        inter_sub_pool = set(dirs) & set(sub_pools)
        pooled_subjects = set(s for p, ss in sub_pools.items() if p in inter_sub_pool for s in ss)
        done_dirs.update(inter_sub_pool)

        ### subjects
        union_sub = set(dirs) | set(subs)
        inter_sub = set(dirs) & set(subs)

        if inter_sub | pooled_subjects == set(subs):
            ok_subs = subs
        else:
            ok_ids = inter_sub | pooled_subjects
            ok_subs = {k:v  for k, v  in subs.items() if k in ok_ids}
            # FIXME not all subjects have folders there may be samples
            # that have folders but not subjects ??? don't wan't to force
            # metadata structure onto folder structure but it complicates
            # the implementation again ... probably worth it in this case
            logd.warning('miscount subject dirs, TODO')

        for subject_id, blob in ok_subs.items():
            if subject_id in pool_subs:  # FIXME assumes 1:1 which is incorrect
                pool_id = pool_subs[subject_id]
            else:
                pool_id = None
            done_dirs.add(subject_id)
            done_specs.add(subject_id)
            records.append({'type': 'SubjectDirs',
                            # have to split the type because we can't recover
                            # the type using just the specimen id (sigh)
                            # and we need it to set the correct prefix (sigh)
                            'specimen_id': subject_id,
                            'dirs': [d[1] for d in dirs[subject_id]]
                            if subject_id in dirs else
                            [d[1] for d in dirs[pool_id]]})

        ### sample pools
        inter_sam_pool = set(dirs) & set(sam_pools)
        pooled_samples = set(s for p, ss in sam_pools.items() if p in inter_sam_pool for s in ss)
        done_dirs.update(inter_sam_pool)

        ### samples
        union_sam = set(dirs) | set(samps)
        inter_sam = set(dirs) & set(samps)

        template_version_less_than_2 = True  # FIXME TODO
        # FIXME this is where non-uniqueness of sample ids becomes a giant pita
        if inter_sam | pooled_samples == set(samps):
            for sample_id, blob in samps.items():
                if sample_id in pool_sams:  # FIXME assumes 1:1 which is incorrect
                    pool_id = pool_sams[sample_id]
                else:
                    pool_id = None

                if len(blob) > 1:
                    # FIXME TODO this means that we need to fail over to the primary keys
                    msg = f'sample_id is not unique! {sample_id}\n{blob}'
                    if he.addError(msg, blame='submission', path=path):
                        logd.error(msg)
                    continue

                if template_version_less_than_2:  # FIXME this is sure the cause an error at some point
                    done_dirs.add((blob[0]['subject_id'], sample_id))
                    done_specs.add(blob[0]['primary_key'])
                    id = blob[0]['primary_key']
                else:
                    done_dirs.add(sample_id)
                    done_specs.add(sample_id)
                    id = sample_id  # FIXME need ttl export suport for this

                records.append({'type': 'SampleDirs',
                                # have to split the type because we can't recover
                                # the type using just the specimen id (sigh)
                                # and we need it to set the correct prefix (sigh)
                                'specimen_id': id,
                                'dirs': [d[1] for d in dirs[sample_id]]
                                if sample_id in dirs else
                                [d[1] for d in dirs[pool_id]]})
        else:
            logd.warning('miscount sample dirs, TODO')
            bad_dirs = []
            if template_version_less_than_2:
                # handle old aweful nonsense
                # 1. construct subject sample lookups using tuple
                # 2. try to construct subject sample id pairs
                for sample_id, blobs in samps.items():
                    for blob in blobs:
                        if sample_id in dirs:
                            candidates = dirs[sample_id]
                            # TODO zero candidates error
                            actual = []
                            for level, drp, rparts in candidates:
                                if level < 2:
                                    msg = (f'Bad location for specimen folder! {drp}')
                                    if he.addError(msg,
                                                   blame='submission',
                                                   path=path):
                                        logd.error(msg)
                                    bad_dirs.append(dirs.pop(sample_id))
                                    continue
                                p_sample_id, p_subject_id, *p_rest = rparts
                                if level < 3:
                                    # p_subject_id will be primary derivatie or source
                                    log.warning(f'TODO new structure {drp}')

                                assert sample_id == p_sample_id  # this should always be true
                                subject_id = blob['subject_id']
                                if subject_id == p_subject_id:
                                    id = blob['primary_key']
                                    done_dirs.add((subject_id, p_sample_id))
                                    done_specs.add(id)
                                    actual.append(drp)

                            if actual:
                                records.append(
                                    {'type': 'SampleDirs',
                                    # have to split the type because we can't recover
                                    # the type using just the specimen id (sigh)
                                    # and we need it to set the correct prefix (sigh)
                                    'specimen_id': id,
                                    'dirs': actual,
                                    })
                        else:
                            msg = f'No folder for sample {sample_id}'
                            if he.addError(msg, blame='submission', path=path):
                                logd.error(msg)
            else:
                pass  # TODO that's an error!

        # handle nesting where parents may not have separate data
        # and this is why we want all ids to be unique per dataset
        # XXX at the moment performance metadata does not require a list of participating specimens
        # and we would have to infer it from the file system hierarchy which is not good
        # TODO populations and friends also an issue here
        _combined_index = {**perfs, **subs, **{v['primary_key']:v for vs in samps.values() for v in vs}}
        # XXX the real solution here is to add a column to the metadata sheets which
        # just makes it possible to assert that there this is a metadata only field
        # which I thought we had already discussed
        def getmaybe(blob):
            # TODO populations are a bit tricky here
            if 'sample_id' in blob:  # FIXME sigh missing types on these
                yield blob['subject_id']
                if 'was_derived_from' in blob:  # TODO member_of works backward with populations, so may need to deal with that case as well
                    parent = blob['was_derived_from']
                    if parent in _combined_index:
                        yield from getmaybe(_combined_index[parent])  # XXX watch out for bad circular refs here
                    else:
                        if parent.startswith('='):
                            msg = f'Sample was_derived_from specified as a formula. We do not currently support this.'
                        else:
                            msg = 'Sample was_derived_from does not exist!'

                        msg += f' {blob["sample_id"]!r} -> {parent!r}'

                        if he.addError(msg, blame='submission', path=path):
                            logd.error(msg)

            elif 'subject_id' in blob:
                return
            elif 'performance_id' in blob:
                return # XXX see note about non-required spec id above
            else:
                raise TypeError(f'unknown blob types {blob}')

        # for all done specs
        # check if there is a parent spec
        # and add it to the maybe done spec list
        maybe_done_specs_all = set(m for i in done_specs | inter_perf for m in getmaybe(_combined_index[i]) if m)
        maybe_not_done_specs = maybe_done_specs_all - set(done_specs)

        usamps = set(v['primary_key'] for vs in samps.values() for v in vs)
        # udirs is a set of strings and tuples, strings for top level
        # folders that are not nested and tuples for samples folders,
        # matching how done_dirs is populated above
        _udirs = set(
            nv for path_name, matched_subpaths in dirs.items() for nv in
            (((subject, path_name) for subject in
              set(element for depth, path, elements in matched_subpaths
                  for element in elements if element.startswith('sub-')))
             if path_name in samps else
             (path_name,)))
        udirs = _udirs  # set(dirs)
        not_done_specs = (set(subs) | usamps) - set(done_specs)
        not_done_dirs = set(udirs) - set(done_dirs)

        def_not_done_specs = not_done_specs - maybe_not_done_specs

        obj = {}

        if records:
            obj['records'] = records
        else:
            pass # TODO embed an error

        if not_done_perfs:
            msg = ('There are performances that have no corresponding '
                   f'directory!\n{not_done_perfs}')
            if he.addError(msg,
                           blame='submission',
                           path=path):
                logd.error(msg)

        if maybe_not_done_specs:
            # FIXME producing incorrect error here when there is a folder but NOT metadata
            # this is strange, this is happening because there is a reference to a subject
            # in a samples sheet that is not in the subjects sheet, that is a separate error
            # FIXME this also means that there is some asymmetry between matching folders
            # and matching specimens because the samples only sheet accounts for the
            # folder but the folder does not account for subjects appearing only in the
            # samples sheet subject_id column
            msg = ('Warning. Unmarked metadata-only specimens that have no corresponding '
                   f'directory, but have children that do.\n{maybe_not_done_specs}')
            if he.addError(msg,
                           blame='submission',
                           path=path):
                logd.warning(msg)

        if def_not_done_specs:
            msg = ('There are specimens that have no corresponding '
                   f'directory!\n{def_not_done_specs}')
            if he.addError(msg,
                           blame='submission',
                           path=path):
                logd.error(msg)

        if not_done_dirs:
            dists = {
                d: sorted(
                    [(nml.levenshteinDistance(d, s), s, source)
                     for ss, source in
                     (((def_not_done_specs | not_done_perfs), ' '),
                      ((done_specs | inter_perf), '*'),)
                     for s in ss])
                for d in not_done_dirs}
            align = {
                d: (lambda l: max(l) if l else -1)(
                    [len(s) for _, s, _ in dss])
                for d, dss in dists.items()}
            report = '\n\n'.join(
                [f'{" ".join([e[1].as_posix() for e in dirs[d]])}\n  0 {d}\n' + '\n'.join(
                    [f'{dist:>3} {s:<{align[d]}} {source}' for dist, s, source in
                     # unfortunately we can't limit the total number of entries because
                     [(dist, s, so) for dist, s, so in dists[d]
                      if dist < 10 or so == ' ' and dist < 15]])
                 for d in sorted(not_done_dirs)])

            msg = ('There are directories that have no corresponding '
                   f'metadata record!\n{not_done_dirs}')
            msg_report = msg + f'\nreport:\n{report}'
            if he.addError(msg + msg_report,
                           blame='submission',
                           path=path):
                logd.error(msg)

        he.embedErrors(obj)
        return obj,

    @staticmethod
    def study_description(study):
        contents = []
        for key in ('study_purpose', 'study_data_collection', 'study_primary_conclusion'):
            if key in study:
                contents.append(study[key])
        out = ' '.join(contents)
        return out,



#!/usr/bin/env python3
# [[file:../../docs/developer-guide.org::combine.py][combine.py]]
import json
import rdflib
from pathlib import Path
from dateutil import parser as dateparser
from pyontutils.core import OntResPath
from pyontutils.utils import TZLOCAL, timeformat_friendly
from pyontutils.namespaces import TEMP, rdf, sparc
#from sparcur.utils import fromJson, register_all_types  # FIXME this should also go in sparcron
from sparcur.export.triples import TriplesExportSummary
from sparcur.export.published import _merge_graphs
from sparcur.simple.export import add_prov

tu = 'timestamp_updated'
tuc = 'timestamp_updated_contents'
ip = 'inputs'
rmk = 'remote_dataset_metadata'


class TriplesExportSummaryPublished(TriplesExportSummary):

    @property
    def ontid(self):
        return rdflib.URIRef(super().ontid.replace('ontologies/', 'ontologies/published/'))

    @property
    def header_label(self):
        return rdflib.Literal(f'{self.folder_name} curation export published graph')


def max_dataset_or_contents_updated(datasets_list):
    return max(set.union(
        set([a['meta'][tuc] for a in datasets_list
            if tuc in a['meta'] and a['meta'][tuc]]),
        set([a['meta'][tu] for a in datasets_list
            if tu in a['meta'] and a['meta'][tu]])))


def from_dataset_export_path_snapshot(dataset_export_path, snapshots_path, time_now):
    derefs = [l.resolve() for l in [c / 'LATEST' for c in dataset_export_path.children] if l.exists()]
    snapshot_path = snapshots_path / time_now.START_TIMESTAMP_LOCAL_FRIENDLY
    snapshot_path.mkdir(parents=True)
    [(snapshot_path / d.parts[-2]).symlink_to((snapshot_path / 'asdf').relative_path_to(d)) for d in derefs]
    return snapshot_path


def from_snapshot_path_datasets_lists(snapshot_path):
    alld = []
    pubd = []
    for uuid_path in snapshot_path.children:
        with open(uuid_path / 'curation-export.json', 'rt') as f:
            j = json.load(f)
            # TODO validate the load XXX this should probably
            # happen as a final roundtrip check during export
            # TODO filter by organization
            alld.append(j)
            if ip in j and rmk in j[ip] and 'id_published' in j[ip][rmk]:
                pubd.append(j)

    return alld, pubd


def from_snapshot_path_summary_json(snapshot_path, project_id, time_now):
    l_all, l_pub = from_snapshot_path_datasets_lists(snapshot_path)
    sum_all = from_datasets_list_summary_json(l_all, project_id, time_now)
    sum_pub = from_datasets_list_summary_json(l_pub, project_id, time_now)
    return sum_all, sum_pub


def from_snapshot_path_summary_ttl(snapshot_path, project_id, time_now, blob):
    tes = TriplesExportSummary(blob)
    tesp = TriplesExportSummaryPublished(blob)

    orps = [OntResPath(uuid_path / 'curation-export.ttl')
            for uuid_path in sorted(snapshot_path.children, key=lambda p: p.name)]

    graphs_all = [o.graph for o in orps]
    graphs_pub = [
        g for g, doi in [(g, list(g[ds:TEMP.hasDoi]))
                         for g in graphs_all
                         for ds in g[:rdf.type:sparc.Dataset]]
        if doi]
    merged_all = _merge_graphs(graphs_all)
    merged_all.populate_from_triples(tes.triples_header)
    merged_pub = _merge_graphs(graphs_pub)
    merged_pub.populate_from_triples(tesp.triples_header)
    return merged_all, merged_pub, graphs_all, graphs_pub

def from_snapshot_path_summary_ttl_BAD(snapshot_path, project_id, time_now, blob):
    # this variant is too complex, trying to reuse the published graph as the all graph
    # and the implementation of the OntConjunctiveGraph is not far enough along to do it
    tes = TriplesExportSummary(blob)  # FIXME nasty annoying dep
    graph_meta = OntGraph()
    graph_meta.populate_from_triples(tes._triples_header(tes.ontid, time_now._start_time))
    rev_replace_pairs = _fix_for_pub(graph_meta, graph_meta)
    replace_pairs = tuple([(b, a) for a, b in rev_replace_pairs])

    orps = [OntResPath(uuid_path / 'curation_export.ttl')
            for uuid_path in snapshot_path.children]

    graphs = [o.graph for o in orps]
    # FIXME should use id_published here as well, but that isn't being
    # propagated at the moment
    graphs_pub = []
    graphs_nop = []
    for g, doi in [(g, list(g[ds:TEMP.hasDoi]))
                   for g in graphs for ds in g[:rdf.type:sparc.Dataset]]:
        if doi:
            graphs_pub.append(g)
        else:
            graphs_nop.add(g)
    graph_pub = _merge_graphs(published_graphs)
    graph_pub.populate_from(graph_meta)
    # FIXME this is manually aligned with TriplesExport.triples_header
    graph_pub.asdf
    for g in graphs_nop:
        graph_pub.namespace_manager.populate_from(
            {k:v for k, v in dict(g.namespace_manager).items()
            if k not in ('contributor', 'sample', 'subject')})

    ttl_all = None
    ttl_pub = _populate_published(curation_export, graph)


def from_dataset_export_path_datasets_lists(dataset_export_path):
    dep = dataset_export_path
    alld = []
    pubd = []
    derefs = [l.resolve() for l in [c / 'LATEST' for c in dep.children] if l.exists()]
    # TODO consider creating a folder that is just symlinks before this
    for lp in sorted(derefs, key=lambda p: p.name):
        with open(lp / 'curation-export.json', 'rt') as f:
            j = json.load(f)
            # TODO validate the load XXX this should probably
            # happen as a final roundtrip check during export
            # TODO filter by organization
            alld.append(j)
            if ip in j and rmk in j[ip] and 'id_published' in j[ip][rmk]:
                pubd.append(j)

    return alld, pubd


def from_datasets_list_summary_json(datasets_list, project_id, time_now):
    # XXX FIXME issue with datasets from multiple projects
    fn = Path(datasets_list[0]['prov']['export_project_path']).name
    out = {
        'id': project_id.id,
        'meta': {
            'count': len(datasets_list),
            'folder_name': fn,  # WHAT A RELIEF we don't need network
            'uri_api': project_id.uri_api,
            'uri_human': project_id.uri_human(),
        },
        'datasets': datasets_list,
    }
    add_prov(out, time_now)
    # some potential prov summaries, but lets not do that here
    # most prov stats should be on the single dataset level
    #'export_timestamp_start_min': min(tes),
    #'export_timestamp_start_max': max(tes),
    return out


def from_dataset_export_path_summary_json(dataset_export_path, project_id, time_now):
    l_all, l_pub = from_dataset_export_path_datasets_lists(dataset_export_path)
    #[a['meta']['timestamp_updated'] < a['meta']['timestamp_updated_contents']
    #for a in l_all if a['meta']['timestamp_updated_contents']]
    sum_all = from_datasets_list_summary_json(l_all, project_id, time_now)
    sum_pub = from_datasets_list_summary_json(l_pub, project_id, time_now)
    return sum_all, sum_pub


def main(project_id=None, export_path=None, time_now=None,
         project_id_auth_var='remote-organization',  # FIXME move to clifun
         disco=False, **kwargs):
    from sparcur.paths import Path
    from sparcur.config import auth

    if project_id is None:
        from sparcur.config import auth
        from sparcur.utils import PennsieveId
        project_id = auth.get(project_id_auth_var)
        project_id = PennsieveId(project_id)  # FIXME abstract the hardcoded backend

    if export_path is None:  # XXX see issues mentioned above
        export_path = Path(auth.get_path('export-path'))

    dataset_export_path = export_path / 'datasets'
    snapshots_path = export_path / 'snapshots'
    snapshot_path = from_dataset_export_path_snapshot(
        dataset_export_path, snapshots_path, time_now)
    sum_all, sum_pub = from_snapshot_path_summary_json(
        snapshot_path, project_id, time_now)
    # write symlink LATEST_PARTIAL
    ttl_all, ttl_pub, graphs_all, graphs_pub = from_snapshot_path_summary_ttl(
        snapshot_path, project_id, time_now, sum_all)
    # write symlink LATEST
    maxdt = max_dataset_or_contents_updated(sum_all['datasets'])
    dt_maxdt = dateparser.parse(maxdt)
    dt_maxdt_local = dt_maxdt.astimezone(TZLOCAL())
    friendly_maxdt_local = timeformat_friendly(dt_maxdt_local)
    # FIXME there are some bad assumptions in here that should be refactored out
    # at some point, but for now we implicitly assume that all datasets come from
    # the same organization, which can easily be violated because we don't check 
    # however the existing internal schema requires an id for the summary which is
    # currenty the organization id
    # FIXME summary is a hardcoded path
    # XXX WARNING it is possible to overwrite since maxdt might not change between runs
    # this is desirable behavior for development, but could cause issues in other cases
    pexpath = export_path / 'summary' / project_id.uuid
    latest = pexpath / 'LATEST'
    npath = pexpath / friendly_maxdt_local
    snapshot_link = npath / 'snapshot'
    if not npath.exists():
        npath.mkdir(parents=True)
    else:
        # FIXME not sure if this makes sense?
        if snapshot_link.is_symlink():
            snapshot_link.unlink()

    snapshot_link.symlink_to(snapshot_link.relative_path_to(snapshot_path))

    npath_ce = npath / 'curation-export.json'
    npath_cep = npath / 'curation-export-published.json'
    for path, blob in ((npath_ce, sum_all),
                       (npath_cep, sum_pub)):
        with open(path, 'wt') as f:
            json.dump(blob, f, indent=2)

    npath_ttl = npath / 'curation-export.ttl'
    npath_ttlp = npath / 'curation-export-published.ttl'
    ttl_all.write(npath_ttl)
    ttl_pub.write(npath_ttlp)

    if disco:
        # export xml and tsv for disco
        from sparcur.simple.disco import from_curation_export_json_path_datasets_json_xml_and_disco
        from_curation_export_json_path_datasets_json_xml_and_disco(
            npath_ce, sum_all['datasets'], graphs_all)

    if latest.is_symlink():
        latest.unlink()

    latest.symlink_to(friendly_maxdt_local)

    return npath_ce, sum_all, sum_pub, graphs_all, graphs_pub


if __name__ == '__main__':
    #import pprint
    from sparcur.simple.utils import pipe_main
    # these are really big, don't print them
    # pipe_main(main, after=pprint.pprint)
    pipe_main(main)
# combine.py ends here



from celery import Celery

_none = 0
_qed = 1
_run = 2
_qed_run = 3

state_lut = {
    _none: 'idle',
    _qed: 'queued',
    _run: 'running',
    _qed_run: 'running-queued',
}


def get_redis_conn():
    rc = Celery(backend='redis://',
                broker='redis://')
    return rc.backend.client


if __name__ == 'sparcur.sparcron':
    import sys
    if (sys.argv[0].endswith('celery') or
        'celery' in sys.argv):
        import sparcur.sparcron.core as celery



import unittest
from .common import skipif_no_net, skipif_ci


@skipif_ci
@skipif_no_net
class TestCron(unittest.TestCase):

    def test_import(self):
        from sparcur import sparcron
        from sparcur.sparcron import core

    def test_sheet_update(self):
        from sparcur.sparcron import core as sparcron
        sparcron.check_sheet_updates()



import copy
import pickle
import unittest
import pytest
import idlib
from sparcur.utils import BlackfynnId, PennsieveId
from idlib.streams import HelpTestStreams


class TestBlackfynnId(unittest.TestCase):

    _id_class = BlackfynnId
    uuids = (('e4d16d59-c963-4d9c-af2f-2e40853881c3', 'package'),)
    cases = (
                     'package:e4d16d59-c963-4d9c-af2f-2e40853881c3',
                    'N:package:e4d16d59-c963-4d9c-af2f-2e40853881c3',
        'https://api.blackfynn.io/packages/N:package:e4d16d59-c963-4d9c-af2f-2e40853881c3',
        'https://api.blackfynn.io/packages/N:package:e4d16d59-c963-4d9c-af2f-2e40853881c3/',
        'https://api.blackfynn.io/packages/N:package:e4d16d59-c963-4d9c-af2f-2e40853881c3/files/1222508',
        'https://api.blackfynn.io/packages/N:package:e4d16d59-c963-4d9c-af2f-2e40853881c3/files/1222508/',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/datasets/N:dataset:fce3f57f-18ea-4453-887e-58a885e90e7e/overview',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/datasets/N:dataset:834e182d-b52c-4389-ad09-6ec9467f3b55/viewer/N:package:a44040e7-5d30-4930-aaac-3aa238ea9081',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/datasets/N:dataset:fce3f57f-18ea-4453-887e-58a885e90e7e/files/N:collection:5bf942a5-10e4-414e-bba6-1f41b053675e',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/datasets/N:dataset:fce3f57f-18ea-4453-887e-58a885e90e7e/files/lol/N:package:457b1339-ac9c-4232-a73e-6c39b1cc1572',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/teams/N:team:d296053d-91db-46ae-ac80-3c137ea144e4',
'https://app.blackfynn.io/N:organization:618e8dd9-f8d2-4dc4-9abb-c6aaab2e78a0/teams/N:team:d296053d-91db-46ae-ac80-3c137ea144e4/',
    )

    def test_regex(self):
        compiled = self._id_class.compiled
        [x.match(u).groups()
           for x, u in ((compiled[x][0], i)
                        for x, i in zip((0,1,3,3,3,3,4,4,4,4,4,4,),
                                        self.cases))
         if not print(u) and not print(x.match(u).groups())]

    def test_uuid(self):
        ids = []
        for uuid, type in self.uuids:
            id = self._id_class(uuid, type=type)
            ids.append(id)

    def test_id(self):
        ids = []
        for string in self.cases:
            id = self._id_class(string)
            ids.append(id)

    @pytest.mark.skip('TODO')
    def test_roundtrip(self):
        # TODO need some way to get/store other component identifiers
        # but tricky when there are 3 identifiers in a single uri
        humans = [case for case in self.cases if 'app.' in case]
        for id_str in humans:
            id = self._id_class(id_str)
            assert id.id in id.uri_human()

    def test_fail_rx(self):
        # TODO bads with edge cases
        try:
            self._id_class('lol not an bfid')
            assert False, 'should have failed'
        except idlib.exc.MalformedIdentifierError as e:  # FIXME malformed id error?
            pass

    def test_pickle(self):
        thing = self._id_class(self.cases[0])
        hrm = pickle.dumps(thing)
        tv = pickle.loads(hrm)
        assert tv == thing

    def test_copy(self):
        thing = self._id_class(self.cases[0])
        thing_prime = copy.deepcopy(thing)
        assert thing_prime == thing

    def test_asCell(self):
        thing = self._id_class(self.cases[0])
        ac = thing.asCell()


class TestPennsieveId(TestBlackfynnId):

    _id_class = PennsieveId
    cases = tuple([c.replace('blackfynn', 'pennsieve') for c in TestBlackfynnId.cases])


@pytest.mark.skip('TODO, need merge of idlib and augpathlib')
class TestIdlibPennsieveId(HelpTestStreams, unittest.TestCase):
    stream = PennsieveId
    ids = TestPennsieveId.cases



from sparcur.sparcron import get_redis_conn
from .endpoints import make_app

conn = get_redis_conn()
app, *_ = make_app(conn)

if __name__ == '__main__':
    app.run(host='localhost', port=7252, threaded=True)



import os
import uuid
import atexit
import shutil
from tempfile import gettempdir
from pathlib import PurePosixPath
import pytest
from augpathlib import PathMeta
from augpathlib.utils import onerror_windows_readwrite_remove
from sparcur import config
from sparcur import exceptions as exc
from sparcur.paths import Path, PathL
from sparcur.paths import LocalPath, PrimaryCache
from sparcur.paths import SymlinkCache
from sparcur.state import State
from sparcur.utils import PennsieveId, log
from sparcur.datasets import DatasetDescriptionFile
from sparcur.curation import PathData, Integrator
from sparcur.pennsieve_api import FakeBFLocal
this_file = Path(__file__).resolve()  # ARGH PYTHON ARGH NO LOL BAD PYTHON
examples_root = this_file.parent / 'examples'
template_root = this_file.parent.parent / 'resources/DatasetTemplate'
print(template_root)
_pid = os.getpid()
path_project_container = this_file.parent / f'test_local-{_pid}'
project_path = PathL(path_project_container) / 'test_project'
fake_org_uuid4 = '0000dead-f00d-4bad-beef-00f000000000'
fake_organization = f'N:organization:{fake_org_uuid4}'
PennsieveId(fake_organization)  # canary
project_path_real = path_project_container / 'UCSD'
test_organization = 'N:organization:ba06d66e-9b03-4e3d-95a8-649c30682d2d'
test_dataset = 'N:dataset:aa859fe9-02d0-4518-981b-012ef8f35c34'
temp_path = Path(gettempdir(), f'.sparcur-testing-base-{_pid}')

onerror = onerror_windows_readwrite_remove if os.name == 'nt' else None

SKIP_NETWORK = ('SKIP_NETWORK' in os.environ or
                'FEATURES' in os.environ and 'network-sandbox' in os.environ['FEATURES'])
skipif_no_net = pytest.mark.skipif(SKIP_NETWORK, reason='Skipping due to network requirement')
skipif_ci = pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')

ds_roots = (
    'ds1',
    'ds2/ds2',
    'ds3/oops',
    'ds3/too',
    'ds3/many',
    'ds4/really/deep',
    'ds5',
    'ds5/multi',
    'ds5/level/wat',
)
ds_folders = set(Path(d).parts[0] for d in ds_roots)

DIVISION_SLASH = '\u2215'  # what have we done XXX evil


def mk_fldr_meta(fldr_path, ftype='collection', id=None):
    tuuid = 'deadbeef' + str(uuid.uuid4())[8:]
    altid = f'N:{ftype}:{tuuid}'  # ok, we validate without remote in a different way
    _id = id if id is not None else altid
    _id = _id.replace('/', DIVISION_SLASH)  # ARGH there must be a way to ensure file system safe ids :/ XXX evil
    _meta = fldr_path.meta
    kwargs = {**_meta}
    kwargs['id'] = _id
    meta = PathMeta(**kwargs)
    # NOTE st_mtime -> modified time of the file contents (data)
    # NOTE st_ctime -> changed time of the file status (metadata)
    # linux does not have a unified way to bet st_btime aka st_crtime which is birth time or created time
    return meta.as_xattrs(prefix='bf')
    #return {'bf.id': kwargs['id'],
            #'bf.created': meta.created.isoformat().replace('.', ',') if meta.created is not None else None,
            #'bf.updated': meta.updated.isoformat().replace('.', ',')}


def mk_file_meta(fp):
    meta = fp.meta
    return meta.as_xattrs(prefix='bf')
    #return {'bf.id': 'N:package:' + fp.as_posix(),
            #'bf.file_id': 0,
            #'bf.size': meta.size,
            # FIXME timezone
            #'bf.created': meta.created.isoformat().replace('.', ',') if meta.created is not None else None,
            #'bf.updated': meta.updated.isoformat().replace('.', ','),
            #'bf.checksum': fp.checksum(),
            # 'bf.old_id': None  # TODO
    #}


def mk_required_files(path, suffix='.csv'):
    # TODO samples.* ?!??!
    # FIXME because empty files get treated as non-existent
    # we either need to treat them as a different category of error
    # OR we need to fill in some fake values when testing
    for globpath in ('submission.*', 'dataset_description.*',  'subjects.*'):
        for template_file in template_root.glob(globpath):
            file_name = template_file.name
            file_path = path / file_name
            shutil.copy(template_file, file_path)
            file_path.touch()
            attrs = mk_file_meta(file_path)
            file_path.setxattrs(attrs)


if project_path.exists():
    # too much time wasted over stale test data
    # like ACSF just make it fresh every time
    project_path.rmtree(onerror=onerror)

project_path.mkdir(parents=True)
atexit.register(lambda : path_project_container.rmtree(onerror=onerror))
attrs = mk_fldr_meta(project_path, 'organization', id=fake_organization)
project_path.setxattrs(attrs)
# FIXME I know why lddi is on cache but I still don't like it
# because it should be possible to call lddi without forcing the cache to exist
# file under CachPath and RemotePath implementations are broken because instances
# can only exist if the underlying thing exists, which defeats the point of paths
project_path._cache_class._asserted_anchor = project_path.cache  # XXX sigh, here we are again
project_path.cache.local_data_dir_init()

for ds in ds_folders:
    dsp = project_path / ds
    dsp.mkdir()
    dsp.setxattrs(mk_fldr_meta(dsp, 'dataset'))

for root in ds_roots:
    rp = project_path / root
    if not rp.exists():
        current_parent = rp
        to_reverse = []
        while not current_parent.exists():
            to_reverse.append(current_parent)
            current_parent = current_parent.parent

        for folder in reversed(to_reverse):
            folder.mkdir()
            folder.setxattrs(mk_fldr_meta(folder))

    print(rp)
    mk_required_files(rp)  # TODO all variants of missing files


fbfl = FakeBFLocal(project_path.cache.id, project_path.cache)
State.bind_blackfynn(fbfl)
#Integrator.setup()  # not needed for tests it seems


@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class RealDataHelper:

    _fetched = False
    _cache_class = None
    _remote_class = None

    @classmethod
    def setUpClass(cls):
        # child classes need to set _cache_class and _remote_class and then super
        from sparcur.paths import Path
        from sparcur.config import auth
        from sparcur.simple import pull, fetch_metadata_files, fetch_files
        test_datasets_real = auth.get_list('datasets-test')
        nosparse = [
            'N:dataset:bec4d335-9377-4863-9017-ecd01170f354',  # mbf headers
        ]
        test_datasets_real.extend(nosparse)
        slot = auth._pathit('{:user-cache-path}/sparcur/objects-temp')
        # FIXME slot needs to be handled transparently as an LRU cache
        # that has multiple folder levels and stores only by uuid
        # and probably lives in '{:user-cache-path}/sparcur/objects'
        slot = slot if slot.exists() else None
        cls.organization_id = auth.get('remote-organization')
        cls.Remote = cls._remote_class._new(Path, cls._cache_class)
        if (hasattr(cls.Remote, '_api') and
            not isinstance(cls.Remote._api, cls.Remote._api_class)):
            log.warning(f'stale _api on remote {cls.Remote._api}')
            for cls in self.Remote.mro():
                if hasattr(cls, '_api'):
                    try:
                        del cls._api
                    except AttributeError as e:
                        pass

        cls.Remote.init(cls.organization_id)
        cls.anchor = cls.Remote.smartAnchor(path_project_container)
        cls.anchor.local_data_dir_init(symlink_objects_to=slot)
        cls.project_path = cls.anchor.local
        list(cls.anchor.children)  # side effect to retrieve top level folders
        datasets = list(cls.project_path.children)
        cls.test_datasets = [d for d in datasets if d.cache_id in test_datasets_real]
        [d.rmdir() for d in datasets if d.cache_id not in test_datasets_real]  # for sanity
        if not RealDataHelper._fetched:
            RealDataHelper._fetched = True  # if we fail we aren't going to try again
            [d._mark_sparse() for d in cls.test_datasets if d.cache_id not in nosparse]  # keep pulls fastish
            pull.from_path_dataset_file_structure_all(cls.project_path, paths=cls.test_datasets)
            fetch_metadata_files.main(cls.project_path)
            fetch_files.main(cls.project_path)


class RDHPN(RealDataHelper):

    @classmethod
    def setUpClass(cls):
        from sparcur.paths import PennsieveCache
        from sparcur.backends import PennsieveRemote
        cls._cache_class = PennsieveCache
        cls._remote_class = PennsieveRemote
        super().setUpClass()
